import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import numpy as np
import html
from time import sleep
import warnings

# Streamlit imports
import streamlit as st
from streamlit import session_state as st_state
from streamlit_option_menu import option_menu
from st_click_detector import click_detector
from st_oxview import oxview_from_text
from code_editor import code_editor

### pyFuRNAce modules
from . import main_menu_style, second_menu_style, copy_to_clipboard
import pyfurnace as pf
from utils.commands import *

code_editor_buttons = ({
                        "name": "Copy",
                        "feather": "Copy",
                        "hasText": True,
                        "alwaysOn": True,
                        "commands": ["copyAll"],
                        "style": {"top": "0.46rem", "right": "0.4rem"}
                        },
                        {
                        "name": "Run",
                        "feather": "Play",
                        "primary": True,
                        "hasText": True,
                        "showWithIcon": True,
                        "commands": ["submit"],
                        "style": {"bottom": "0.44rem", "right": "0.4rem"}
                        },
                        )

funny_bootstrap_icons = ['robot', 'trash', 'umbrella', 'camera', 'cart', 'cpu', 
                         'cup-straw', 'trophy', 'palette', 'cup-straw', 
                         'camera-reels', 'puzzle', 'hourglass-split', 'mortarboard']

direction_list = pf.Direction.names()

def origami_general_options(origami, expanded=True):
    with st.expander('General settings', expanded=expanded):
        cols = st.columns(3)

        ### select alignment
        with cols[0]:
            align = st.radio("Alignment type:", 
                             ['To the left', 
                             'Junctions alignment: first', 
                             'Line center'], 
                             index=0, 
                             key='alignment')
                
            ### add the alignment to the code
            if st.button('Set alignment', key='set_alignment'):
                align = align.split()[-1]
                origami.align = align
                st_state.code.append(f"origami.align = '{align}' # Align the lines")

        with cols[1]:
            st.toggle('Optimize the blueprint for ROAD', 
                      value=False,
                      key='to_road', 
                      help='Optimize the blueprint for the ROAD software. '
                      'This substitutes the Kissing Loops base pairings with "*";'
                      ' and the short stem base pairings with "!".')

        with cols[2]:
            new_sticky = st.toggle('Stick the motif menu at the top', 
                                   value=st_state.motif_menu_sticky, 
                                   key='sticky_menu', 
                                   help='Keep the motif menu and origami visualization'
                                   ' menu to stick to the top of the page.')
            if new_sticky != st_state.motif_menu_sticky:
                st_state.motif_menu_sticky = new_sticky
                st.rerun()

        col1, col2 = st.columns(2)
        with col1:
            font_size = st.slider('Origami font size', 
                                  min_value=2, 
                                  max_value=50, 
                                  value=14, 
                                  key='ori_font_size')
            st_state.origami_font_size = font_size

        with col2:
            frame_size = st.slider('Oxview frame size (disable and renable '
                                   'the OxView to apply changes)',
                                    min_value=0, 
                                    max_value=2000, 
                                    value=500, 
                                    key='oxFrame_size')
            st_state.oxview_frame_size = frame_size

def simple_origami():
    with st.form(key='simple_origami_form'):
        dt_text = st.text_input("Enter a list of the angles between helices, "
                                "separated by commas", 
                                "120, ", 
                                key='dt_text_list',
                                help= f"The program calculates the best connections"
                                    f" bewteen the helices to fit the given angles."
                                    f"The connection between helices (Dovetails) are "
                                    f"obtained roughly with this lookup table "
                                    f"(angle --> dt): {pf.ANGLES_DT_DICT}")

        angle_list = [int(x) for x in dt_text.split(",") if x and x.strip()]
        dt_list = pf.convert_angles_to_dt(angle_list)
        main_stem_default = 11 * ((max([abs(dt) for dt in dt_list], 
                                        default=0) + 17) // 11 + 1)
        col1, col2 = st.columns(2, vertical_alignment='bottom')
        with col1:
            kl_columns = st.number_input('Kissing loop columns:', 
                                         min_value=1, 
                                         value=1, 
                                         help='number of KL repeats in the helix')
        with col2:
            main_stem = st.number_input('Spacing between crossovers (bp):', 
                                        min_value=22, 
                                        value=main_stem_default, 
                                        step=11, 
                                        help='The length of the consecutive stems '
                                              'in the helix')

        submitted = st.form_submit_button("Submit")
        if submitted:
            st_state.origami = pf.simple_origami(dt_list=angle_list, 
                                                         kl_columns=kl_columns, 
                                                         main_stem=main_stem, 
                                                         add_terminal_helix=True, 
                                                         align=st_state.origami.align,
                                                         use_angles=True)
            st_state.code.append(f'origami = pf.simple_origami(dt_list={angle_list}, '
                                 f'kl_columns={kl_columns}, '
                                 f'main_stem={main_stem}, '
                                 f'add_terminal_helix=True, '
                                 f'align="{st_state.origami.align}", use_angles=True) '
                                 f'# Create a simple origami')
            
            # select the end of the origami
            st_state.line_index = len(st_state.origami) - 1
            st_state.motif_index = len(st_state.origami[-1])

def motif_text_format(motif: pf.Motif) -> str:
    # take a motif string and expand it on right, left, top and bottom with spaces
    motif_list = ([[' '] * (motif.num_char + 2)]
                    + [[' '] + [char for char in line] + [' '] 
                                        for line in str(motif).split('\n')] 
                    + [[' '] * (motif.num_char + 2)]
                    )
    
    ### Add the 5' and 3' to the motif text
    for s in motif:

        if not s.sequence: # skip strand without sequence
            continue

        y0, x0 = s.prec_pos[1] + 1, s.prec_pos[0] + 1
        if s[0] not in '35' and motif_list[y0][x0] == ' ':
            motif_list[y0][x0] = '<' + s.directionality[0] + '>'

        y1, x1 = s.next_pos[1] + 1, s.next_pos[0] + 1
        if s[-1] not in '35' and motif_list[y1][x1] == ' ':
            motif_list[y1][x1] = '<' + s.directionality[1] + '>'

    # remove the top and bottom lines if they are empty
    for i in (0, -1):
        if all([char == ' ' for char in motif_list[i]]): 
            motif_list.pop(i)

    motif_str = '\n'.join([''.join(line) for line in motif_list])
        
    preview_txt = (motif_str.replace(' ', '&nbsp;')
                   .replace('\n', '<br />')
                   .replace('<5>', '<span style="color: #D52919;">5</span>')
                   .replace('<3>', '<span style="color: #D52919;">3</span>'))
    return preview_txt

def initiate_session_state():
    if 'origami' not in st_state:
        st_state.origami = pf.Origami()
    if "code" not in st_state:
        st_state.code = ["import pyfurnace as pf", 
                         "origami = pf.Origami()"]
        st_state.motif_buffer = ""
        st_state.mod_motif_buffer = ""
    if "motif" not in st_state:
        st_state.motif = pf.Motif()
    # if "redo" not in st_state:
    #     st_state.redo = []
    if 'motif_menu_sticky' not in st_state:
        st_state.motif_menu_sticky = False
    if "copied_motif" not in st_state:
        st_state.copied_motif = None
    if "copied_motif_text" not in st_state:
        st_state.copied_motif_text = ""
    if 'modified_motif_text' not in st_state:
        st_state.modified_motif_text = ""
    if 'upload_key' not in st_state:
        st_state.upload_key = 0
    ### Initialize the session state variables for the custom motif
    if 'custom_strands' not in st_state:
        st_state.custom_strands = []
    if 'custom_motifs' not in st_state:
        st_state.custom_motifs = [('Custom Motif', 
                                   'dpad', 
                                   pf.Motif(lock_coords=False))]
    if 'custom_key' not in st_state:
        st_state.custom_key = 0
    if "custom_edit" not in st_state:
        st_state["custom_edit"] = True
    if "last_clicked_position" not in st_state:
        st_state["last_clicked_position"] = ''
    if 'line_index' not in st_state:
        st_state.line_index = 0
    if 'motif_index' not in st_state:
        st_state.motif_index = 0
    if 'current_line_occupied' not in st_state:
        st_state.current_line_occupied = False
    if 'ori_click_count' not in st_state:
        st_state.ori_click_count = 0
    if 'max_pk_index' not in st_state:
        st_state.max_pk_index = 1
    if 'oxview_selected' not in st_state:
        st_state.oxview_selected = ()
    if 'selected_motif' not in st_state:
        st_state.selected_motif = 'Connections'
    if 'gradient' not in st_state:
        st_state.gradient = False
    if 'oxview_colormap' not in st_state:
        st_state.oxview_colormap = 'Reds'
    if 'origami_font_size' not in st_state:
        st_state.origami_font_size = 14
    if 'oxview_frame_size' not in st_state:
        st_state.oxview_frame_size = 500

def update_file_uploader():
    st_state.upload_key += 1

@st.fragment
def make_motif_menu(origami):
    # update max pk index
    pk_motifs = origami[lambda m: hasattr(m, 'pk_index')]
    st_state.max_pk_index = max([abs(int(x.pk_index.replace("'", ""))) 
                                        for line in pk_motifs 
                                            for x in line], 
                                default=0) + 1

    option_data = {'Connections': 'bi-sliders',
                   'Structural': 'bi-bricks',
                   'Kissing Loops': 'bi-heart-half',
                   'Aptamers': 'bi-palette',
                   'Custom': 'bi-joystick',
                   'Edit': 'bi-bandaid',
                   'Code': 'bi-code'}
    
    select_ind = 0
    if st_state.selected_motif in option_data:
        select_ind = list(option_data).index(st_state.selected_motif)        
    
    selected_motif = option_menu(None, 
                                 list(option_data.keys()),
                                 icons=list(option_data.values()),
                                 menu_icon="cast", 
                                 orientation="horizontal",
                                 manual_select=select_ind,
                                 styles=main_menu_style,
                                 key='motif_menu')
    
    if selected_motif != st_state.selected_motif:
        st_state.selected_motif = selected_motif
        try:
            st.rerun(scope='fragment')
        except Exception as e:
            print(f"Error in rerun: {e}")
            st.rerun()

    motif_add = True

    if selected_motif == 'Connections':
        ConnectionsCommand().execute()
    elif selected_motif == 'Structural':
        StructuralCommand().execute()
    elif selected_motif == 'Kissing Loops':
        KissingLoopsCommand().execute()
    elif selected_motif == 'Aptamers':
        AptamersCommand().execute()
    elif selected_motif == 'Custom':

        ### Adding the menu here, otherwise there are issues choosing the 
        # custom motif with st.fragment when the edit mode is off
        col1, col2 = st.columns([5, 1], vertical_alignment='bottom')
        with col1:
            motif_selected = option_menu(None,
                                        [l[0] for l in st_state.custom_motifs],
                                        icons=[f"bi bi-{l[1]}" 
                                               for l in st_state.custom_motifs],
                                        menu_icon="cast", 
                                        orientation="horizontal",
                                        styles=second_menu_style)
        with col2:
            new_name = st.text_input(":green[Add custom motif with name:]", 
                                     key=f'new_custom_motif{st_state.custom_key}')
            if new_name:
                icon = funny_bootstrap_icons[0]
                # rotate the icons
                funny_bootstrap_icons.append(funny_bootstrap_icons.pop(0))
                st_state.custom_motifs.append((new_name, 
                                               icon, 
                                               pf.Motif(lock_coords=False)))
                st_state.custom_key += 1
                st.rerun()

        current_custom_motif = [m for m in st_state.custom_motifs 
                                        if m[0] == motif_selected][0][2]
        custom(current_custom_motif)

    elif selected_motif == 'Edit':
        motif_add = False
        edit(st_state.motif_index, st_state.line_index)

    elif selected_motif == 'Code':
        code()
        motif_add = False
    # elif selected_motif == 'Undo/Redo':
    #     _, col1, _, col2, _ = st.columns([1] * 5)
    #     with col1:
    #         undo()
    #     with col2:
    #         redo()
    #     motif_add = False

    if motif_add:
        st.markdown("<hr style='margin-top:-0em;margin-bottom:-1em' />", 
                    unsafe_allow_html=True)
        add_motif(origami)

def select_line(f_col1=None, f_subcol2=None, f_subcol3=None):
    warnings.filterwarnings("error") # raise warnings as errors
    origami = st_state.origami
    origami_len = len(origami)
    line_index = st_state.line_index
    motif_index = st_state.motif_index
    clicked_indexes = st_state.get(f"origami_click{st_state.ori_click_count}")

    if line_index < 0:
        line_index = 0
    elif motif_index < 0:
        motif_index = 0

    ### Check if the origami is empty and add an empty line
    if origami_len == 0:
        origami.append([]) # add empty line if the Origami is empty
        st_state.code.append('origami.append([]) # Add empty line')
        origami_len = 1
    
    col1, col2, col3, col4, col5, col6 = st.columns([1, 1.5, 1.5, 1, 1, 1], 
                                                    vertical_alignment='bottom')

    with col1:
        if f_col1: f_col1()

    ### Select the line in which to add the motif
    with col2:
        line_index = min(line_index, origami_len - 1)
        line_index = st.number_input("**Select line index:**", 
                                    min_value=-1,
                                    max_value=origami_len - 1, 
                                    value=line_index, 
                                    key=f'select_line_index{st_state.ori_click_count}')
        st_state.current_line_occupied = False

        if line_index < origami_len and origami[line_index]:
            st_state.current_line_occupied = True
        if line_index != st_state.line_index:
            st_state.line_index = line_index
            if clicked_indexes:
                st_state.ori_click_count += 1
            st.rerun()

        subcol1, subcol2 = st.columns(2)
        
        with subcol1:
            if f_subcol2:
                f_subcol2(line_index + 1)

        ### delete line
        with subcol2:
            # check that the origami has more than one line, or the line is not empty
            if len(st_state.origami) > 1 or st_state.origami[0]: 
                # If we want to add a new line, we stop the function here
                if st_state.line_index >= len(st_state.origami):
                    return 
                if len(st_state.origami) > 0:
                    del_line = st.button("Delete line", 
                                         key = 'del_line', 
                                         help="Delete the choosen line")
                    if del_line:
                        st_state.origami.pop(st_state.line_index) #remove choosen helix
                        st_state.code.append(f'origami.pop({st_state.line_index})')
                        st_state.line_index -= 1
                        st.rerun() # rerun app
            # if there is the button to add a motif, add a empty button to align
            elif f_subcol3: 
                st.button('', type='tertiary')

    if st_state.line_index >= len(origami):
        return

    ### Motif index selection
    with col3:
        max_val = len(origami[line_index])
        if motif_index > max_val:
            motif_index = max_val
        motif_index = st.number_input("**Select motif index:**", 
                                      min_value=0, 
                                      max_value=max_val, 
                                      value=motif_index, 
                                key = f'select_motif_index_{st_state.ori_click_count}')
        
        if motif_index != st_state.motif_index:
            st_state.motif_index = motif_index
            if clicked_indexes:
                st_state.ori_click_count += 1
            st.rerun()
        subcol1, subcol2 = st.columns(2)

        with subcol1:
            if f_subcol3:
                f_subcol3()
        with subcol2:
            ### Delete motif
            if len(origami[st_state.line_index]) > 0:
                delete_button = st.button(':red[Delete motif]', key = 'delete_motif')
                if delete_button:
                    if st_state.motif_index == len(origami[st_state.line_index]):
                        st_state.motif_index -= 1
                    origami.pop((st_state.line_index, st_state.motif_index))
                    st_state.code.append(f'origami.pop(({st_state.line_index}, '
                                         f'{st_state.motif_index})) # Delete motif')
                    # st_state.redo = [] # clear the redo buffer
                    st.rerun() # rerun app

    ### duplicate the current line
    with col4:
        if st.button("Duplicate line", key='duplicate_line'):
            # warnings.filterwarnings("ignore") # ignore kl energy warning
            origami.duplicate_line(line_index, insert_idx=len(origami))
            st_state.code.append(f'origami.duplicate_line({line_index}, '
                                 f'insert_idx={len(origami)}) # Duplicate line')
            st_state.line_index = len(origami) - 1
            st.rerun() # rerun app

    ### copy/paste motif
    with col5:
        subcol1, subcol2 = st.columns(2)
        with subcol1:
            copy_motif()
        # Paste motif
        with subcol2:
            paste_button = st.button('Paste motif', key = 'paste_motif')
            if paste_button:
                if not st_state.copied_motif:
                    st.warning('No motif copied')
                    return
                origami.insert((line_index, motif_index), st_state.copied_motif)
                st_state.code.append(st_state.copied_motif_text + 
                                     f'\norigami.insert(({line_index}, '
                                     f'{motif_index}), motif) # Paste motif')
                # st_state.redo = []
                st.rerun()

    # Uno action
    with col6:
        undo(key='motif_undo')


def add_motif(origami):
    ### Load the motif and prepare the motif buffer for custom motifs
    motif = st_state.motif
    if not motif:
        return
    
    # If the motif is a generic motif, generate the motif text buffer
    elif st_state.selected_motif == 'Custom': 
        ### generate the motif code
        st_state.motif_buffer = "strands = []\n"
        for s in st_state["custom_strands"]:
            st_state.motif_buffer += (f"strands.append(pf.Strand('{s}', "
                                     f"directionality='{s.directionality}', "
                                     f"start={tuple(s.start)}, "
                                     f"direction={tuple(s.direction)}, "
                                     f"coords=pf.Coords({s.coords.array.tolist()}, "
                                     f"dummy_ends=({s.coords.dummy_ends[0].tolist()}, "
                                     f"{s.coords.dummy_ends[1].tolist()}))))\n")
        
        st_state.motif_buffer += (f"motif = pf.Motif(strands, "
                                  f"structure = '{motif.structure}', "
                                  f"lock_coords={motif.lock_coords})")

    ### Show the preview of the motif
    def f_col1():
         # write Preview in small and in orange 
        st.write(':orange[Preview:]')
        # display the Motif to be added next
        scrollable_text(motif_text_format(motif))

    ### add line
    def f_subcol2(linex_index):
        # check that the origami has more than one line, or the line is not empty
        if len(st_state.origami) > 1 or st_state.origami[0]: 
            add_line = st.button("Add line", 
                                 key='Insert_line', 
                                 help=f"Add a new line n° {linex_index}.")
            if add_line:
                #add empty line at the choosen position
                origami.insert(linex_index, []) 
                st_state.code.append(f'origami.insert({linex_index}, [])') 
                st_state.line_index += 1
                st.rerun()

    ### Add motif
    def f_subcol3():
            button_label = 'Insert motif'
            if st_state.motif_index == len(origami[st_state.line_index]):
                button_label = 'Add motif'
            add_button = st.button(f'**:green[{button_label}]**', 
                                   key='insert_motif', 
                                   help="Insert the motif at the choosen position")
            
            if add_button:
                origami.insert((st_state.line_index, st_state.motif_index), motif) 
                st_state.code.append(st_state.motif_buffer + 
                                     f'\norigami.insert(({st_state.line_index}, '
                                     f'{st_state.motif_index}), motif) # Add motif')
                st_state.motif_buffer = '' ### clear the buffer
                # st_state.redo = [] # clear the redo buffer
                st_state.motif_index += 1
                st.rerun()

    select_line(f_col1, f_subcol2, f_subcol3)


def copy_motif(key='', motif=None, motif_slice=None):
    copy_button = st.button('Copy motif', key = f'copy_motif{key}')
    if copy_button:
        if not motif:
            st_state.copied_motif = st_state.motif
            st_state.copied_motif_text = st_state.motif_buffer
        else:
            st_state.copied_motif = motif
            st_state.copied_motif_text = (f"motif = origami[{motif_slice}].copy() "
                                          f"# Copy motif at line {motif_slice[0]}, "
                                          f"index {motif_slice[1]}")

###
# TAB FUNCTIONS
###    

@st.fragment
def generate_custom_motif_text(strand, x_size=50, y_size=10):
    ### update the content
    content = (f"<div style='font-family: monospace; "
               f"font-size: {st_state['origami_font_size'] + 8}px;'>")
    current_motif = st_state.motif
    m_pos_to_sym = current_motif.get_position_map()
    s_pos_to_sym = strand.get_position_map()

    for y in range(y_size):
        for x in range(x_size):
            if (x, y) in s_pos_to_sym:
                content += (f'<a href="javascript:void(0);" id="{x},{y}" '
                            f'style="color: #D00000;">{s_pos_to_sym[(x, y)]}</a>')
                
            elif (x, y) in m_pos_to_sym: # strand not currently selected
                symbol = m_pos_to_sym[(x, y)]
                content += (f'<a href="javascript:void(0);" id="{x},{y}" '
                            f'style="color: #00856A; opacity: 0.5;">{symbol}</a>')
                
            else:
                content += (f'<a href="javascript:void(0);" id="{x},{y}" '
                            f'style="color: grey; opacity: 0.5;">•</a>')
        content += '<br />'
    content += "</div>"
    return content

def custom_text_input(current_custom_motif):
    # make a motif list adding a space around the motif
    motif_lines = str(current_custom_motif).split('\n')
    add_line = False
    add_char = False
    if current_custom_motif:
        if any(line[0] != ' ' for line in motif_lines):
            add_char = True

        motif_list = [[' '] * add_char + [char for char in line] + [' '] 
                                                        for line in motif_lines]

        if motif_lines[0].strip():
            add_line = True
            motif_list = [[' '] * (current_custom_motif.num_char + 2)] + motif_list
        if motif_lines[-1].strip():
            motif_list += [[' '] * (current_custom_motif.num_char + 2)]

    else:
        motif_list = [['']]

    # Add the  5 before the start of each motif.
    for s in current_custom_motif:

        ys, xs = s.prec_pos[1] + int(add_line), s.prec_pos[0] + int(add_char)
        ye, xe = s.next_pos[1] + int(add_line), s.next_pos[0] + int(add_char)
        # add 5' to the start of the strand
        if (s.directionality == '53' 
            and motif_list[ys][xs] == ' '): 
            motif_list[ys][xs] = '5'
         
        # add 5' to the end of the strand
        elif (s.directionality == '35' 
              and motif_list[ye][xe] == ' '):
            motif_list[ye][xe] = '5'

    current_custom_motif_str = '\n'.join(''.join(line) for line in motif_list)

    strand_text = st.text_area("Motif text: draw a motif where each Strand starts "
                               "with a 5", 
                               value=current_custom_motif_str, 
                               key="Motif_text", 
                               help='Draw a motif, where each strand has to start '
                                    'with "5". If you want to start a strand with 5, '
                                    'add an additional 5 at the beginning of the '
                                    'strand.'
                                )
    
    if strand_text != current_custom_motif_str:
        if '5' not in strand_text:
            st.warning('Don\'t forget to start the strand with "5"')
        else:
            new_motif = pf.Motif.from_text(strand_text).strip()
            current_custom_motif.replace_all_strands(new_motif._strands, 
                                                     copy=False)
            current_custom_motif.basepair = new_motif.basepair
            st.rerun()

def structure_converter(current_custom_motif):
    ### Add the structure converter
    col1, col2 = st.columns(2)
    with col1:
        structure = st.text_input("Dot-bracket structure:", 
                                 value=current_custom_motif.structure,
                                 key="Dot-bracket_structure", 
                                 help="Add a dot-bracket structure to convert it into "
                                      "a 3D motif.")
    with col2:
        default_seq = str(current_custom_motif.sequence)
        if len(structure) > len(default_seq):
            default_seq += 'N' * (len(structure) - len(default_seq))
        sequence = st.text_input("Sequence:", 
                                 value=default_seq, 
                                 key="Sequence", 
                                 help="Add the sequence of the motif.")
        if not sequence:
            sequence = None
    if (structure and structure != current_custom_motif.structure or
            sequence and sequence != current_custom_motif.sequence):
        try:
            new_motif = pf.Motif.from_structure(structure, sequence=sequence)
        except Exception as e:
            st.error(f"Error: {e}")
            return
        current_custom_motif.replace_all_strands(new_motif._strands, copy=False)
        current_custom_motif.basepair = new_motif.basepair
        st.rerun()

def upload_3d_interface(strand, strand_num, current_custom_motif):
    file_3d = st.file_uploader(f"3D coordinates (OxDNA format) of "
                               f"**strand {strand_num}**", 
                               type=['dat', 'pdb'],
                               key=f'custom_{st_state.upload_key}', 
                               help='Upload an Oxview configuration ".dat" '
                                    'file with the 3D coordinates of one strand.')
    dummy_cols = st.columns(2)

    dummy_help = ('The dummy base is a base in the coordinates that is not '
                  'part of the sequence, but it is used to connect other strands '
                  'to the beginning or end of the strand.')

    with dummy_cols[0]:
        dummy_start = st.toggle('Start dummy nucleotide', 
                                key=f'dummy_start_custom', 
                                help=dummy_help)
    with dummy_cols[1]:
        dummy_end = st.toggle('End dummy nucleotide', 
                              key=f'dummy_end_custom', 
                              help=dummy_help)
        
    if file_3d:
        pdb_format = file_3d.name.endswith('.pdb')
        strand.coords = pf.Coords.load_from_text(file_3d.getvalue().decode('utf-8'),
                                                dummy_ends=(dummy_start, dummy_end),
                                                pdb_format=pdb_format)
        update_strand(strand, 
                      strand_num, 
                      current_custom_motif, 
                      coords=True)

def update_strand(strand, strand_num, current_custom_motif, coords=None):
    st_state["custom_strands"][strand_num] = strand
    current_custom_motif.replace_all_strands(st_state["custom_strands"], 
                                             copy=False)
    if coords:
        st.success("3D coordinates uploaded successfully!", 
                   icon=":material/view_in_ar:")
        sleep(2)
        update_file_uploader()
    st.rerun()

def strand_number_button(current_custom_motif, delete=True):
    strands = [s.copy() for s in st_state["custom_strands"]]
    strand_num = st.radio("Select strand:", 
                          list(range(len(strands))), 
                          index=len(strands)-1)
    strand = strands[strand_num]

    if (delete 
            and st.button("Delete strand", key="delete_strand") 
            and current_custom_motif):
        
        st_state["custom_strands"].pop(strand_num)
        current_custom_motif.pop(strand_num)
        if not st_state["custom_strands"]:
            st_state["custom_strands"].append(pf.Strand(''))
        st.rerun()

    return strand, strand_num


@st.fragment
def custom(current_custom_motif):
    ### Silence the warnings
    warnings.filterwarnings("ignore")
    st_state.motif = current_custom_motif

    if not st_state["custom_edit"]:
        if st.button(":orange[Edit the motif]", key="edit_strand"):
            st_state["custom_edit"] = True
            st.rerun()
        return
    
    GeneralEditCommand().execute(current_custom_motif)
    
    st_state["custom_strands"] = [s.copy() for s in current_custom_motif]
    if not current_custom_motif:
        st_state["custom_strands"] = [pf.Strand('')]

    col1, col2 = st.columns([4, 1], vertical_alignment='bottom')
    with col1:
        method = st.segmented_control('Build the motif with:',
                                    ['Structure conveter',
                                    'Drawing Tool', 
                                    'Full text input',
                                    ],
                                    default='Structure conveter',
                                    help='Structure conveter: Convert a dot-bracket'
                                    ' structure into a 3D motif automatically.\n\n'
                                    'Drawing Tool: Draw each strand on a grid by '
                                    'clicking on the positions where you want the '
                                    'strand to go; curves and crossings are '
                                    'calculated automatically.\n\n'
                                    'Full text input: Create a motif on a blank text'
                                    ' area by typing the motif strand '
                                    'and basepairing.\n'
                                    )
        
    with col2:
        ### add instructions and common symbols
        with st.popover("Common symbols to copy"):
            common_pyroad_sym = ["─", "│", "╭", "╮", "╰", "╯", "^", "┼", '┊', "&"]
            cols = st.columns(len(common_pyroad_sym))
            for i, sym in enumerate(common_pyroad_sym):
                with cols[i]:
                    copy_to_clipboard(sym, sym)

    if method == 'Structure conveter':
        structure_converter(current_custom_motif)
    elif method == 'Full text input':
        custom_text_input(current_custom_motif)
    if method != 'Drawing Tool':

        with st.popover("Upload 3D coordinates"):
            subcol1, subcol2 = st.columns([1, 4])
            with subcol1:
                strand, strand_nr = strand_number_button(current_custom_motif, 
                                                         delete=False)
            with subcol2:
                upload_3d_interface(strand, strand_nr, current_custom_motif)

        if current_custom_motif:
            st.write('Current motif preview:')
            scrollable_text(motif_text_format(current_custom_motif))
            finish_editing(current_custom_motif)
        return

    cols = st.columns(4, vertical_alignment='bottom')
    with cols[0]:
        if st.button("Add strand", key="add_strand"):
            st_state["custom_strands"].append(pf.Strand(''))
            current_custom_motif.append(pf.Strand(''), copy=False, join=False)
    with cols[1]:
        x_dots = st.number_input("Canvas x size", min_value=1, value=64, key="x_size")
    with cols[2]:
        y_dots = st.number_input("Canvas y size", min_value=1, value=8, key="y_size")
    with cols[3]:
        if st.button("Clear", key="clear_strand"):
            st_state["custom_strands"] = [pf.Strand('')]
            current_custom_motif.replace_all_strands([pf.Strand('')], copy=False)

    # Update and display the motif text
    col1, col2 = st.columns([2, 7])
    with col1:
        strand, strand_num = strand_number_button(current_custom_motif)
    with col2:
        try:
            clicked = click_detector(generate_custom_motif_text(strand, 
                                                    x_size=x_dots, 
                                                    y_size=y_dots), 
                                                    key="custom_motif_click_detector")
        except Exception as e:
            st.error(str(e))

    def move_strand(clicked, add_nucl=None):
        nonlocal strand
        pos = tuple([int(i) for i in clicked.split(',')])

        ### if the strand is empty, add the first position
        if not str(strand):
            strand._start = pos
            strand.strand = "─"

        ### if the strand has only one position, add the strand direction
        elif len(str(strand)) == 1:
            start = strand.start
            direction = (pos[0] - start[0], pos[1] - start[1])
            if direction[0] and direction[1]:
                st.error("Invalid strand, the strand can't form diagonal lines.")
                return
            # the Start horizontally
            if direction[0]:
                sym = '-'
                distance = (abs(direction[0]) + 1)
                if direction[0] > 0:
                    strand.direction = (1, 0)
                else:
                    strand.direction = (-1, 0)
            # the Start vertically
            else:
                sym = '|'
                distance = (abs(direction[1]) + 1)
                if direction[1] > 0:
                    strand.direction = (0, 1)
                else:  
                    strand.direction = (0, -1)
            if add_nucl:
                sym = add_nucl
            strand.strand = sym * distance
        else:
            ### if the clicked position is the start, remove it and update the start
            if pos == strand.start:
                strand._start = strand.positions[0]
                strand._direction = strand.directions[0]
                strand.strand = strand.strand[1:]

            ### check the direction and in case add the new position
            else:
                # calculate the direction
                direction = (pos[0] - strand.end[0], pos[1] - strand.end[1])
                # calculate the direction symbol
                if add_nucl:
                    sym = add_nucl
                elif direction[0]:
                    sym = "─"
                else:
                    sym = "│"
                # calulcate the difference between the new direction and 
                # the end direction (use to determine the symbol to add)
                normal_direction = tuple([i // abs(i) if i != 0 else 0 
                                                            for i in direction])
                direction_difference = (normal_direction[0] - strand.end_direction[0],
                                        normal_direction[1] - strand.end_direction[1])
                distance = max(abs(pos[0] - strand.end[0]), 
                               abs(pos[1] - strand.end[1]))
                # if the direction isn't changed, add the last symbol 
                # to match the strand
                if normal_direction == strand.end_direction:
                    strand.strand = strand.strand + sym * distance
                elif normal_direction not in [(1, 0), (-1, 0), (0, 1), (0, -1)]:
                    st.error("Invalid strand. The strand can't form diagonal lines.")
                elif direction_difference == (-1, 1):
                    strand.strand = strand.strand[:-1] + "╮" + sym * distance
                elif direction_difference == (1, 1):
                    strand.strand = strand.strand[:-1] + "╭" + sym * distance
                elif direction_difference == (-1, -1):
                    strand.strand = strand.strand[:-1] + "╯" + sym * distance
                elif direction_difference == (1, -1):
                    strand.strand = strand.strand[:-1] + "╰" + sym * distance
                else:
                    st.error("Invalid strand")

    ### Update the strand on click
    if clicked and clicked != st_state["last_clicked_position"]:
        try:
            move_strand(clicked)
            st_state["last_clicked_position"] = clicked
        except pf.MotifStructureError as e:
            st.error(str(e))

    ### Show the strand options
    col1, col2, col3, col4, col5 = st.columns([1, 1, 1, 1, 5])
    with col1:
        start_x = st.number_input("Start x:", 
                                  min_value=0, 
                                  value=strand.start[0], 
                                  key=f'start_x_custom')
    with col2:
        start_y = st.number_input("Start y:", 
                                  min_value=0, 
                                  value=strand.start[1], 
                                  key=f'start_y_custom')
    with col3:
        strand_direction_ind = [d for d in pf.Direction].index(strand.direction)
        new_dir = st.selectbox('Start direction:', 
                               direction_list, 
                               index=strand_direction_ind, 
                               key=f'dir_custom')
        new_dir_tuple = pf.Direction[new_dir]
    with col4:
        seq_dir = st.selectbox('Directionality:', 
                               ['35', '53'], 
                               index=['35', '53'].index(strand.directionality), 
                               key=f'seq_dir_custom')
    with col5:
        new_strand = st.text_input(f'New strand (strand directionality: '
                                   f'{strand.directionality}) ', 
                                   value=str(strand), key=f'strand_custom')
    
    ### update the strand
    strand.start = (start_x, start_y)
    strand.direction = new_dir_tuple
    strand.strand = new_strand
    strand.directionality = seq_dir
    with st.popover('Add strand 3D coordinates'):
        upload_3d_interface(strand, strand_num, current_custom_motif)

    # compare the strand with the version strand: if the strand is different,
    #  update it and rerun
    old_strand = st_state["custom_strands"][strand_num]
    if strand != old_strand:
        update_strand(strand, strand_num, current_custom_motif)

    ### check the base pair symbols of the motif
    current_structure = current_custom_motif.structure
    new_db = st.text_input('Add dot-bracket notation:', 
                           value=current_structure,  
                           key=f'structure_custom', 
                           help='Add the dot-bracket notation of the motif for each '
                                'strand, separated by a "&". If the paired bases are '
                                'more than one position apart, the pairing symbol "┊"'
                                ' is not shown.')
    if new_db != current_structure:
        current_custom_motif.structure = new_db

    finish_editing(current_custom_motif)
    st.stop()

def finish_editing(current_custom_motif):
    if st.button(':green[Finish editing to add motif to the origami]', 
                 key="end_edit"):
        st_state["custom_edit"] = False
        current_custom_motif.replace_all_strands(st_state["custom_strands"], 
                                                 copy=False)
        st.rerun()

def update_code(code_text):
    # Check for forbidden keywords
    forbidden_keywords = ['exec', 'eval', 'compile', ' open', ' os', ' sys', ' __']
    for kw in forbidden_keywords:
        if kw in code_text:
            st.error(f"Forbidden keyword detected in the code: '{kw}'", 
                     icon=":material/sentiment_extremely_dissatisfied:")
            return False
    
    # Define the local environment in which the user's code will be executed
    local_context = {'origami': st_state.origami}
    # Attempt to execute the code safely
    # try:
    exec(code_text, {'__builtins__': __builtins__, 'pf': pf}, local_context)
     # Retrieve the modified origami variable
    st_state.origami = local_context['origami'] 
    st.success("Nanostructure updated successfully!")

    # select the end of the origami
    st_state.line_index = len(st_state.origami) - 1
    st_state.motif_index = len(st_state.origami[-1])

    # except Exception as e:
    #     st.error(f"Error in executing the code: {e}")
    #     return False
    st_state.code = code_text.split('\n\n')
    st.rerun()

def code():
    if "last_code_id" not in st_state:
        st_state["last_code_id"] = ''
    code_text = '\n\n'.join(st_state.code)

    render_lines = st.slider("Number of lines to render:", 
                             min_value=1, 
                             max_value=200, 
                             value=15, 
                             key="render_lines")

    response_dict = code_editor(code_text, 
                                buttons=code_editor_buttons,
                                height=render_lines,
                                options={"showLineNumbers":True}, 
                                allow_reset=True, 
                                key="pyroad_code_editor")
    
    if (response_dict['id'] != st_state["last_code_id"] 
            and (response_dict['type'] == "submit" 
                 or response_dict['type'] == "selection") ):
        st_state["last_code_id"] = response_dict['id']
        update_code(response_dict['text'])

    else:
        st.success("Structure updated!")

def undo(key=''):
    if not st_state.origami:
        return
    if len(st_state.code) <= 1:
        st.warning("Nothing to undo.")
        return
    undo_button = st.button("Undo", 
                            key=f'undo{key}', 
                            use_container_width=False,
                            help="Undo the last action.")
    if not undo_button:
        return
    # st_state.redo.append(st_state.code[-1])
    update_code('\n\n'.join(st_state.code[:-1]))

# def redo():
#     if not st_state.redo:
#         # st.warning("Nothing to redo.")
#         return
#     redo_button = st.button("Redo", help="Redo the last action.")
#     if not redo_button:
#         return
#     last_action = st_state.redo.pop()
#     update_code('\n\n'.join(st_state.code + [last_action]))
#     st.rerun()

def origami_select_display():
    option_data = {"Origami 2D view": "bi bi-square", 
                   "Origami 3D view": "bi bi-box", 
                   "Origami split view": "bi bi-window-split",
                   'Folding barrier': "bi bi-graph-up"}

    selected_display = option_menu(None,
                                list(option_data.keys()),
                                icons=list(option_data.values()),
                                orientation='horizontal',
                                key='DisplayMenu',
                                styles=main_menu_style
                                )
    return selected_display
    

@st.fragment
def origami_build_view(selected_display):
    ### Display the RNA origami structure with clickable elements
    #  and modify them in case
    warnings.filterwarnings("ignore") # ignore numpy warnings

    if selected_display == 'Origami 2D view':
        clicked = display_origami()
        clicked_options(clicked)

    elif selected_display == 'Origami 3D view':
        display3d()

    elif selected_display == 'Origami split view':
        col1, col2 = st.columns(2)
        with col1:
            clicked = display_origami()
        with col2:
            display3d()
        clicked_options(clicked)

    elif selected_display == 'Folding barrier':
        barriers, score = st_state.origami.folding_barriers()
        clicked = display_origami(barriers=barriers)
        clicked_options(clicked)
        
        help_message = ("The folding barriers are a simplified calculation to check " 
                        "if a structure is feasible to form co-transcriptionally. "
                        "They are calculated considering that pseudoknots lock the "
                        "structure in a specific conformation; preventing the "
                        "formation of successive stems. A barrier happens when "
                        "a stem is open before the formation of a pseudoknots, "
                        "and is closed after the formation of the pseudoknot "
                        "(after a 150-bases delay). The penalty is calculated as W"
                        " (weak barrier): 1 point, S (strong barrier): 2 points, "
                        "and the total penalty is the sum of the points.")

        cols = st.columns(4, vertical_alignment='center')
        with cols[0]:
            if score < 30:
                st.success(f"Low folding barrier penalty: {score}.")
            elif score < 100:
                st.warning(f"Medium folding barrier penalty: {score}.")
            else:
                st.error(f"High folding barrier penalty: {score}.")
        with cols[1]:
            st.markdown("", help=help_message)
        with cols[2]:
            if score > 30:
                if st.button('Try to fix the folding pathway', key='fix_pathway',
                          help="Try to change the starting position of the sequence, "
                               "keeping the same structure, to reduce the folding "
                               "pathway penalty. This function is designed for "
                               "canonical RNA Origami structures, it may not work "
                               "for other complex structures."):
                    
                    # calculate the simplfied dot bracket
                    with st.spinner("Optimizing the folding pathway..."):
                        code_lines = st_state.code
                        code_lines += ['\norigami = origami.improve_''folding_pathway'
                                       '(kl_delay=150)\n']
                        update_code('\n'.join(code_lines))
                    
def display3d():
    origami = st_state.origami
    
    if st_state.gradient:
        seq = origami.sequence.replace('&', '')
        index_colors = list(range(len(seq)))

    else:
        m_slice = (st_state.line_index, st_state.motif_index)
        index_colors = ()
        if m_slice[0] < len(origami) and m_slice[1] < len(origami[m_slice[0]]):
            motif = origami[m_slice[0]][m_slice[1]]
            motif_shift = origami.index_shift_map[m_slice]
            index_colors = [0] * len(origami.sequence.replace('&', ''))
            for pos in motif.seq_positions:
                shifted = tuple([pos[0] + motif_shift[0], pos[1] + motif_shift[1]])
                index_colors[origami.seq_positions.index(shifted)] = 1

    if index_colors:
        for s in origami.strands:
            for protein in s.coords.proteins:
                index_colors += [0] * len(protein)

    conf, topo = origami.save_3d_model("origami", return_text=True)
    oxview_from_text(configuration=conf, # path to the configuration file
                    topology=topo,      # path to the topology file
                    width='99%',                # width of the viewer frame
                    colormap=st_state.oxview_colormap, # colormap for the viewer
                    height=st_state.oxview_frame_size, # height of the viewer frame
                    index_colors=index_colors, # color the bases in the viewer
                    key='display_nano')  

def build_origami_content(barriers=None):
    barriers_colors ={'▂':'#FFBA08', '▄':'#FFBA08', '█':'#D00000'}
    highlight_color = '#D00000'
    normal_color = '#333333' # 80% black

    origami = st_state.origami
    motif = origami.assembled

    if barriers:
        origami_lines = origami.barrier_repr(barriers=barriers, 
                                             return_list=True)
    else:
        if st_state.get('to_road'):
            origami_str = origami.to_road()

        else:
            origami_str = str(origami)
        origami_lines = origami_str.split('\n')

        
    # create color gradient
    if st_state.gradient:
        # create a dictionary from positions to index
        pos_to_index = {pos: ind for ind, pos in enumerate(motif.seq_positions)}
        tot_len = 0
        for s in origami.strands:
            tot_len += len(s.sequence)
            for protein in s.coords.proteins:
                tot_len += len(protein)
        cmap = plt.get_cmap(st_state.oxview_colormap)
        oxview_offset = tot_len // 5
        c_map = [mcolors.to_hex(cmap(i)) 
                            for i in np.linspace(0, 1, tot_len + oxview_offset)]
        c_map = c_map[oxview_offset:]
    
    # Prepare the string to add 5' and 3' symbols for the strands
    motif_list = ([[' '] * (motif.num_char + 2)] 
                  + [[' '] + [char for char in line] + [' '] for line in origami_lines]
                  + [[' '] * (motif.num_char + 2)])
    
    for s in motif:  # Add the 5' and 3' symbols to the motif as 1 and 2
        if not s.sequence:
            continue
        ys, xs = s.prec_pos[1] + 1, s.prec_pos[0] + 1
        ye, xe = s.next_pos[1] + 1, s.next_pos[0] + 1
        # add 5' to the start of the strand
        if (s.sequence and s[0] not in '35' 
                and motif_list[ys][xs] == ' '):
            motif_list[ys][xs] = '1' if s.directionality == '53' else '2'
        # add 5' to the end of the strand
        if (s.sequence and s[-1] not in '35' 
                and motif_list[ye][xe] == ' '):
            motif_list[ye][xe] = '2' if s.directionality == '53' else '1'
    origami_list = [''.join(line) for line in motif_list]

    content = ("<div style='font-family: monospace; white-space: nowrap; "
               "overflow-x: auto;'>")
    
    # add a column with the line number
    content += (f'<div style="display:inline-block; font-family: monospace;'
                f' font-size: {st_state.origami_font_size}px;">Line:<br />' )
    
    line_nr = -2
    origami_list_len = len(origami_list)
    span_text = '<span style="font-family: monospace; '

    for y, line in enumerate(origami_list):
        line_color = normal_color
        motif_slice = None
        current_line_nr = [origami.pos_index_map[(x-1, y-1)] 
                                for x, _ in enumerate(line) 
                                    if (x-1, y-1) in origami.pos_index_map]
        
        next_line = origami_list[y+1] if  (y + 1) < origami_list_len else []
        
        # check the next line index
        next_line_nr = [origami.pos_index_map[(x-1, y)] 
                                for x, _ in enumerate(next_line) 
                                        if (x-1, y) in origami.pos_index_map]

        if current_line_nr: 
            current_line_nr = current_line_nr[0][0]  # get the first line number
        if next_line_nr:
            next_line_nr = next_line_nr[0][0]

        # is a new origami line
        if line_nr != current_line_nr and isinstance(current_line_nr, int):  
            if current_line_nr == st_state.line_index:
                line_color = highlight_color
            content += (span_text 
                        + (f'color: {line_color}; line-height:1;">'
                           f'{current_line_nr})</span>')
                        + (4 - len(str(current_line_nr))) * '&nbsp;' )
            line_nr = current_line_nr

        # the origami line is empty
        elif (isinstance(next_line_nr, int) 
                    and line_nr < st_state.line_index < next_line_nr):
            content += span_text + 'color: #D52919; line-height:1;">_____</span>'

        # is not a new origami line
        else:
            content += (span_text + 
                        'line-height:1;">&nbsp;&nbsp;&nbsp;'
                        '&nbsp;&nbsp;</span>' )
        
        hit = False # to highlight the first symbol
        for x, char in enumerate(line):
            ori_pos = (x - 1, y - 1)
            color = normal_color
            if barriers_colors and char in barriers_colors:
                color = barriers_colors[char]

            if char == ' ':
                content += span_text + 'line-height:1;">&nbsp;</span>'

            elif char == '1':
                content += (span_text 
                            + f'color: {highlight_color}; line-height:1;">5</span>')
            elif char == '2':
                content += (span_text 
                            + f'color: {highlight_color}; line-height:1;">3</span>')
                
            # do not highlight the base pair in red
            elif char in pf.bp_symbols:
                content += (span_text 
                            + f'color: {color}; line-height:1;">{char}</span>' )

            elif ori_pos in origami.pos_index_map:  # a motif symbol
                motif_slice = origami.pos_index_map[ori_pos]
                if st_state.gradient: 
                    index = pos_to_index.get(ori_pos)
                    if index is not None:
                        color = c_map[index]

                # This is the selected motif
                if (motif_slice 
                        and motif_slice[0] == st_state.line_index 
                        and motif_slice[1] == st_state.motif_index):
                    
                    # Don't add color if the gradient is not active
                    if st_state.gradient: 
                        color = normal_color

                    # the first symbol is yellow
                    elif not hit: 
                        color = '#FF8800'
                        hit = True

                    else:
                        color = highlight_color

                content += (f'<a style="font-family: monospace; color: {color}; '
                            'line-height:1;" href="javascript:void(0);" '
                           f'id="{motif_slice[0]},{motif_slice[1]},{x - 1},'
                           f'{y - 1}">{char}</a>')

            # is a junction symbol 
            else:  
                content += span_text + f'color: {color}; line-height:1;">{char}</span>'

        # Check if you wanna add the cursor
        if motif_slice and motif_slice[0] == st_state.line_index:
            if len(origami[motif_slice[0]]) == st_state.motif_index:
                content += (span_text 
                            + f'color: {highlight_color}; line-height:1;">│</span>')
        content += '<br />'

    if line_nr < st_state.line_index:
        # the origami line is empty
        content += span_text + 'color: #D52919; line-height:1;">_____</span>'

    content += '</div>'
    content += '</div>'
    return content


def display_origami(barriers=None):
    ### SHORTCUTS NOT READY YET!!!
    # if not split_view:
    #     def move_selection(x = 0, y = 0):
    #         st_state.line_index += y
    #         st_state.motif_index += x
    #         # st.rerun()

    #     with st.expander("Shortcuts to move over the Origami", expanded=False):
    #         st.write("Use the arrow keys to move the selection over the Origami "
    #                   "lines and motifs. Not available in split view.")
    #         col1, col2, col3, col4 = st.columns(4)
    #         with col1:
    #             streamlit_shortcuts.button("Move Up", on_click=move_selection, 
    #                                           shortcut="Shift+ArrowUp", 
    #                                            hint=True, kwargs={"y" : -1})
    #         with col2:
    #             streamlit_shortcuts.button("Move Down", 
    #                                         on_click=move_selection, 
    #                                         shortcut="Shift+ArrowDown", 
    #                                          hint=True, kwargs={"y" : 1})
    #         with col3:
    #             streamlit_shortcuts.button("Move Left", 
    #                                        on_click=move_selection, 
    #                                         shortcut="Shift+ArrowLeft", 
    #                                         hint=True, kwargs={"x" : -1})
    #         with col4:
    #             streamlit_shortcuts.button("Move Right", 
    #                                         on_click=move_selection, 
    #                                         shortcut="Shift+ArrowRight", 
    #                                         hint=True, kwargs={"x" : 1})

    try:
        content = build_origami_content(barriers=barriers)
        clicked = click_detector(content, 
                                 key=f"origami_click{st_state.ori_click_count}")
    except pf.MotifStructureError as e:
        st.error(f"Structure error: \n {e}", icon=":material/personal_injury:")
        undo('error', key='error')
        st.stop()
    except pf.AmbiguosStructure as e:
        st.error(f"Ambigouse structure: \n {e}", icon=":material/theater_comedy:")
        undo('warning', key='warning')
        st.write("You can try flipping the motif or changing the sequence direction.")
        st.stop()
    return clicked

def clicked_options(clicked):
    # undo()
    # col1, _ , _, col2, _ = st.columns([1] * 5)
    # with col1:
    #     undo()
    # with col2:
    #     redo()

    if clicked:
        y_slice, x_slice, x_pos, y_pos  = [int(i) for i in clicked.split(',')]
        motif_slice = (y_slice, x_slice)
        pos = (x_pos, y_pos)

        # Write the selected motif on screen
        try:
            motif = st_state.origami[motif_slice]
        except IndexError:
            return # this is a problem with updating the keys
        
        motif_class_name = motif.__class__.__name__
        if motif_class_name == 'Motif': ### is custom Motif
            motif_class_name = "Custom " + motif_class_name

        nucl_text = ''
        # Indicate the nucleotide index:
        if pos in st_state.origami.seq_positions:
            ind = st_state.origami.seq_positions.index(pos)
            nucl_text += f':orange[nucleotide index {ind}] in'

        ### highlight the selected motif
        if (st_state.line_index != motif_slice[0] 
                or st_state.motif_index != motif_slice[1]):
            st_state.line_index = motif_slice[0]
            st_state.motif_index = motif_slice[1]
            st.rerun()

        col1, col2, _ = st.columns([1, 1, 1], vertical_alignment='center')
        with col1:
            st.markdown(f'Selected {nucl_text} :orange[{motif_class_name}]: '
                        f':orange[line {motif_slice[0]}], '
                        f':orange[motif {motif_slice[1]}]')
        with col2:
            ### copy the motif
            copy_motif('selected', motif=motif, motif_slice=motif_slice)

def display_structure_sequence():
    st.write("#### Origami structure and sequence constraints:")
    origami = st_state.origami
    if origami.structure:
        structure_list = [char for char in origami.structure]
        sequence_list = [char for char in origami.sequence]
        motif_slice = (st_state.line_index, st_state.motif_index)
        indexes = []
        if motif_slice in origami.index_shift_map:
            motif = origami[motif_slice]
            shifts_x, shift_y = origami.index_shift_map[motif_slice]
            for pos in motif.seq_positions:
                indexes.append(origami.seq_positions.index((pos[0] + shifts_x, 
                                                            pos[1] + shift_y)))

        for i in range(len(structure_list)):
            if i in indexes:  # highlight the selected motif
                structure_list[i] = (f'<span style="color: #D52919; line-height:1;">'
                                    f'{html.escape(structure_list[i])}</span>')
                sequence_list[i] = (f'<span style="color: #D52919; line-height:1;">'
                                    f'{html.escape(sequence_list[i])}</span>')
            else:
                structure_list[i] = html.escape(structure_list[i])
                sequence_list[i] = html.escape(sequence_list[i])
            # structure_list[i] = f":orange[{structure_list[i]}]"
            # sequence_list[i] = f":orange[{sequence_list[i]}]"

        ### Pseudoknots info
        pseudoknot_text = ('; '.join([f'id: {k}, ' + str(val) 
                                        for k, val in origami.pseudoknots.items()]) 
                          + ';')
        remove_syms = str.maketrans('', '', "{}'\"")
        pseudoknot_text = html.escape(pseudoknot_text.translate(remove_syms))

        st.markdown(f'**Structure length: '
                    f':green[{str(len(origami.structure.replace("&", "")))}]**')
        
        scrollable_text(f">Origami</br>{''.join(sequence_list)}</br>"
                        f"{''.join(structure_list)}</br></br>"
                        f"Pseudoknots info:</br>{pseudoknot_text}")

        st.markdown("<hr style='margin-top:-0em;margin-bottom:-1em;"
                    "color:white;border-color:white' />", 
                    unsafe_allow_html=True)

        st_state.generate_structure = origami.structure
        st_state.generate_sequence = str(origami.sequence)
        st_state.generate_pseudoknots = pseudoknot_text
        col1, col2, col3, col4 = st.columns(4)
        with col1:
            st.page_link("pages/2_Generate.py", 
                         label="**:orange[Generate the sequence]**", 
                         icon=":material/network_node:")
        with col2:
            copy_to_clipboard(origami.structure, 'Structure')
        with col3:
            copy_to_clipboard(origami.sequence, 'Sequence')
        with col4:
            copy_to_clipboard(pseudoknot_text, 'Pseudoknots')

def edit(x, y):
    origami = st_state.origami
    ### select the motif to change
    st_state.modified_motif_text = '' # initialize the modification
    try:
        motif_slice = (y, x)
        if y >= len(origami) or x >= len(origami[y]):
            select_line()
            return
        motif = origami[motif_slice]
    except KeyError:
        return # this is a problem with updating the keys
    motif_class_name = motif.__class__.__name__
    if motif_class_name == 'Motif': ### is custom Motif
        motif_class_name = "Custom " + motif_class_name
    elif isinstance(motif, pf.KissingLoop):
        motif_class_name = 'KissingLoops'

    ### Directly modify the motif and make a copy for the advanced feature
    command = f"{motif_class_name}Command"
    if command in globals():
        globals()[command]().execute(motif)
    else:
        GeneralEditCommand().execute(motif)

    if st_state.modified_motif_text:
        st_state.code.append(f'motif = origami[({motif_slice[0]}, '
                             f'{motif_slice[1]})] '
                             f'# motif slice: line {motif_slice[0]}, '
                             f'index {motif_slice[1]}' 
                             + st_state.modified_motif_text)
        st_state.modified_motif_text = ''
        st.rerun()

    select_line()
    st_state['mot_adv_edit'] = motif.copy()
    advaced_edit(motif_slice)

@st.fragment
def advaced_edit(motif_slice):
    y, x = motif_slice

    ### try to change each strand
    with st.popover('Advanced feature, modify the strands of the selected motif:',
                    use_container_width=True,):
        
        cols = st.columns(7, vertical_alignment='bottom')
        with cols[0]:
            flip_vert = st.toggle("Flip vertically", 
                                value=True,
                                key="adv_flip_vert")
        with cols[1]:
            flip_h = st.toggle("Flip horizontally", 
                                value=True,
                                key="adv_flip_h")
        with cols[2]:
            if flip_vert or flip_h:
                flip = st.button("Flip", key=f"adv_flip")
                flip_vert &= flip
                flip_h &= flip
                
        with cols[3]:
            if st.button("Rotate 90° clockwise", key="adv_rotate"):
                st_state.mot_adv_edit.rotate()
                st_state.modified_motif_text += ('\nmotif.rotate()')

        if flip_h or flip_vert:
            st_state.mot_adv_edit.flip(horizontally=flip_h, vertically=flip_vert)
            st_state.modified_motif_text += (f'\nmotif.flip('
                                                f'horizontally={flip_h}, '
                                                f'vertically={flip_vert})')


        with cols[4]:
            x_shift = st.number_input("Horizontal shift amount:", value=0)
        with cols[5]:
            y_shift = st.number_input("Vertical shift amount:", value=0)

        with cols[6]:
            shift = st.button("Shift motif", 
                              help="Shift the motif by the specified x and y values. "
                                   "The shift is applied to all strands in the motif.")

        if (x_shift != 0 or y_shift != 0) and shift:
            st_state.mot_adv_edit.shift((x_shift, y_shift))
            st_state.modified_motif_text += (f'\nmotif.shift(('
                                            f'{x_shift}, {y_shift}))')

        for i, s in enumerate(st.session_state.mot_adv_edit):
            col1, col2, col3, col4, col5 = st.columns([1, 1, 1, 1, 5])
            with col1:
                start_x = st.number_input("Start x:", 
                                            min_value=0, 
                                            value=s.start[0], 
                                            key=f'start_x_{x}_{y}_{i}')
            with col2:
                    start_y = st.number_input("Start y:", 
                                                min_value=0, 
                                                value=s.start[1], 
                                                key=f'start_y_{x}_{y}_{i}')
                    
            with col3:
                strand_direction_ind = [d for d in pf.Direction].index(s.direction)
                new_dir = st.selectbox('Start direction:', 
                                        direction_list, 
                                        index=strand_direction_ind, 
                                        key=f'dir_{x}_{y}_{i}')
                new_dir_tuple = pf.Direction[new_dir]

            with col4:
                seq_dir = st.selectbox('Directionality:', 
                                        ['35', '53'], 
                                        index=['35', '53'].index(s.directionality), 
                                        key=f'seq_dir_{x}_{y}_{i}')

            with col5:
                new_strand = st.text_input(f'New strand (strand directionality: '
                                            f'{s.directionality}) ', 
                                            value=str(s), key=f'strand_{x}_{y}_{i}',
                                            help=' The strand contains all '
                                                'the symbols of the structure '
                                                '(nucleotides and direction '
                                                'symbols). ')


            ### check if the strand is modified and update the motif
            strand_stripped = new_strand.strip()
            if s.start != (start_x, start_y):
                s.start = (start_x, start_y)
                st_state.modified_motif_text += (f'\nmotif[{i}].start = '
                                                    f'({start_x}, {start_y})')
            if s.directionality != seq_dir:
                s.directionality = seq_dir
                st_state.modified_motif_text += (f'\nmotif[{i}].directionality '
                                                    f'= "{seq_dir}"')
            if s.direction != new_dir_tuple:
                s.direction = new_dir_tuple
                st_state.modified_motif_text += (f'\nmotif[{i}].direction = '
                                                    f'{new_dir_tuple}')
            if s.strand != strand_stripped:
                s.strand = strand_stripped
                st_state.modified_motif_text += (f'\nmotif[{i}].strand = '
                                                    f'"{strand_stripped}"')

        ### check the base pair symbols of the motif
        current_structure = st_state.mot_adv_edit.structure
        new_db = st.text_input('Add dot-bracket notation:', 
                                value=current_structure,  
                                key=f'structure_{x}_{y}', 
                                help='Add the dot-bracket notation of the motif for'
                                    ' each strand, separated by a "&". If the '
                                    'paired bases are more than one position '
                                    'apart, the pairing symbol "┊" is not shown.')
        
        if new_db != current_structure:
            if not new_db:
                st_state.mot_adv_edit.autopairing = True
                st_state.modified_motif_text += f'\nmotif.autopairing = True'
            else:
                st_state.mot_adv_edit.structure = new_db
                st_state.modified_motif_text += f'\nmotif.structure = "{new_db}"'
        
        ### update the motif
        try:
            st.write(':orange[Preview:]')

            scrollable_text(motif_text_format(st_state.mot_adv_edit))
            if not st_state.modified_motif_text:
                st.write(':green[Structure is updated]') 
            else:
                st.warning('Be careful before updating, there is a high risk of '
                            'breaking the structure.')
                update = st.button('Update strands')

                if update:
                    st_state.code.append(f'motif = origami[({motif_slice[0]}, '
                                            f'{motif_slice[1]})] # select the motif '
                                            f'at index, line: {motif_slice}' 
                                            + st_state.modified_motif_text)
                    
                    st_state.origami[motif_slice] = st_state.mot_adv_edit.copy()
                    st.rerun()

        except pf.MotifStructureError as e:
            st.error(e)


def scrollable_text(text: str):
        st.markdown(
        f"""
        <style>
        .scroll-box {{
            overflow-x: auto;
            white-space: nowrap;
            background-color: #fafafa;
            padding: 10px;
            border-radius: 20px; /* Rounded corners */
            font-size: {st_state.origami_font_size + 1}px;
        }}
        </style>
        <div class="scroll-box">{text}</div>
        """, unsafe_allow_html=True)


import streamlit as st
from streamlit import session_state as st_state
from .motif_command import MotifCommand

class GeneralEditCommand(MotifCommand):
    
    def execute(self, motif=None):
        ### Modify motif
        if motif:
            flip_vert, flip_hor, rotate = self.interface('mod')
            if flip_vert or flip_hor:
                st_state.modified_motif_text += (f"\nmotif.flip(horizontally="
                                                 f"{flip_hor}, "
                                                 f"vertically={flip_vert})")
                motif.flip(horizontally=flip_hor, vertically=flip_vert)
            elif rotate:
                st_state.modified_motif_text += f"\nmotif.rotate({rotate})"
                motif.rotate(rotate)

    @staticmethod
    def interface(key=''):
        col1, col2, col3 = st.columns(3, vertical_alignment='bottom')
        
        if key:
            with col1:
                subcol1, subcol2 = st.columns(2)
                with subcol1:
                    flip_vert = st.toggle("Flip vertically", 
                                        value=True,
                                        key=f"{key}flip_vert")
                with subcol2:
                    flip_hor = st.toggle("Flip horizontally", 
                                        value=True,
                                        key=f"{key}flip_hor")
            with col2:
                if flip_vert or flip_hor:
                    flip = st.button("Flip", key=f"{key}flip")
                    flip_vert &= flip
                    flip_hor &= flip
            with col3:
                 rotate = st.button("Rotate 90° clockwise", key=f"{key}rotate")
                 
             
        else:
            with col1:
                flip_vert = st.toggle("Flip vertically", key=f"{key}flip_vert")
            with col2:
                flip_hor = st.toggle("Flip horizontally", key=f"{key}flip_hor")
            with col3:
                rotate = st.number_input("Rotate 90° clockwise:", 
                                         min_value=0, 
                                         max_value=4, 
                                         key=f"{key}rotate")

        return flip_vert, flip_hor, rotate



from typing import Union, Tuple, List, Callable
from Bio.Seq import Seq
from Bio.SeqUtils import MeltingTemp as mt
from Bio.SeqUtils import gc_fraction
from functools import partial

# https://www.bioinformatics.org/sms/iupac.html
# dictionary of melting temperature methods: 
# name as key and function as value
tm_methods = {"Nearest Neighbor": mt.Tm_NN,
              "Empirical GC content": mt.Tm_GC,
              "Wallace, rule of thumb": mt.Tm_Wallace,
              }

# dictionary of melting temperature models:
#  name as key and function as value
tm_models = {"Nearest Neighbor": ['DNA_NN4', 
                                  'DNA_NN1', 
                                  'DNA_NN2', 
                                  'DNA_NN3', 
                                  'RNA_NN1', 
                                  'RNA_NN2', 
                                  'RNA_NN3', 
                                  'R_DNA_NN1'],
             "Empirical GC content": [1, 2, 3, 4, 5, 6, 7, 8],
             "Wallace, rule of thumb": [],
             }

# dictionary of Nearest Neighbour melting temperature models:
#  name as key and function as value
NN_models = {"DNA_NN4": mt.DNA_NN4,
             "DNA_NN1": mt.DNA_NN1,
             "DNA_NN2": mt.DNA_NN2,
             "DNA_NN3": mt.DNA_NN3,
             "RNA_NN1": mt.RNA_NN1,
             "RNA_NN2": mt.RNA_NN2,
             "RNA_NN3": mt.RNA_NN3,
             "R_DNA_NN1": mt.R_DNA_NN1
             }

# dictionary of default values for the primer energy parameters

default_values = {"Na": 0, "K": 50, "Tris": 20, "Mg": 1.5,
                  "dNTPs": 0.2, "Method": 7, "DMSO (%)": 0,
                  "mt_method": list(tm_methods)[0],
                  "mt_model": tm_models[list(tm_methods)[0]][0],
                  "Primer": 500,
                  }


def make_tm_calculator(method_name: str = "Nearest Neighbor",
                        model_name: str = "DNA_NN4",
                        primer_conc: float = 500,
                        tm_kwargs: dict = default_values.copy()) -> callable:
    """
    Creates a melting temperature (Tm) calculator function for DNA primers using 
    specified calculation method and model.
    Parameters
    ----------
    method_name : str, optional
        The method used for Tm calculation. 
        Options include "Nearest Neighbor", 
        "Empirical GC content",
        and "Wallace, rule of thumb".
        Default is "Nearest Neighbor".
    model_name : str, optional
        The nearest neighbor model to use for Tm calculation. 
        Only relevant if `method_name` is "Nearest Neighbor".
        Default is "DNA_NN4".
    primer_conc : float, optional
        The concentration of the primer (in nM) used in the calculation. 
        Only relevant for "Nearest Neighbor" method.
        Default is 500.
    tm_kwargs : dict, optional
        Additional keyword arguments for the Tm calculation method. 
        May include chemical corrections (e.g., DMSO, dNTPs).
        Default is a copy of `default_values`.
    Returns
    -------
    callable
        A function that takes a DNA sequence (str) and returns its melting 
        temperature (float), possibly corrected for chemical additives.
    
    Notes
    -----
    - If DMSO is specified in `tm_kwargs`, the returned function 
        applies a chemical correction to the calculated Tm.
    - For "Nearest Neighbor" method, dNTPs and DMSO are removed 
        from the keyword arguments before calculation.
    - For "Empirical formulas based on GC content", 
        DMSO is removed from the keyword arguments before calculation.
    - For "Wallace" method, no additional arguments are processed.
    Examples
    --------
    >>> tm_calc = make_tm_calculator(method_name="Nearest Neighbor", primer_conc=250)
    >>> tm_calc("ATCGATCG")
    62.5
    """
    method = tm_methods[method_name]

    first_word = method_name.split()[0].lower()

    if first_word == "nearest":
        model = NN_models[model_name]
        method_kwargs = tm_kwargs.copy()
        if 'dNTPs' in method_kwargs:
            method_kwargs.pop('dNTPs')
        if 'DMSO' in method_kwargs:
            method_kwargs.pop('DMSO')
        partial_func = partial(method, 
                               nn_table=model, 
                               dnac1=primer_conc, 
                               dnac2=0, 
                               **method_kwargs)

    elif first_word == "empirical":
        method_kwargs = tm_kwargs.copy()
        if 'DMSO' in method_kwargs:
            method_kwargs.pop('DMSO')
        partial_func = partial(method, valueset=model_name, **method_kwargs)

    else:  # Wallace
        partial_func = method

    if tm_kwargs.get('DMSO', 0) > 0:
        return lambda seq: mt.chem_correction(partial_func(seq), 
                                              DMSO=tm_kwargs['DMSO'])
    
    return partial_func


def calculate_gc(seq: str) -> float:
    """
    Calculate the GC content percentage of a nucleotide sequence.

    Parameters
    ----------
    seq : str
        The nucleotide sequence for which to calculate the GC content.

    Returns
    -------
    float
        The GC content of the sequence as a percentage, rounded to one decimal place.

    Notes
    -----
    Ambiguous nucleotides are ignored in the calculation.
    """
    return round(gc_fraction(seq, ambiguous='ignore') * 100, 1)


def annealing_temp(mts: List[float], 
                   seq: str, 
                   tm_kwargs: dict, 
                   method: str = 'IDT') -> float:
    """
    Calculate the annealing temperature for PCR primers using either 
    the IDT or Phusion method.

    Parameters
    ----------
    mts : List[float]
        List of melting temperatures (Tm) for the primers.
    seq : str
        DNA sequence for which the annealing temperature is to be calculated.
    tm_kwargs : dict
        Dictionary of keyword arguments for Tm calculation, such as 
        salt concentration, DMSO, etc.
    method : str, optional
        Method to use for calculation. Either 'IDT' or 'Phusion'. 
        Default is 'IDT'.

    Returns
    -------
    float
        Calculated annealing temperature, rounded to two decimal places.

    Notes
    -----
    - The 'IDT' method applies a chemical correction and a weighted average formula.
    - The 'Phusion' method uses the minimum Tm, with an adjustment for primer length.
    - Certain keys in `tm_kwargs` (e.g., 'dNTPs', 'DMSO') are handled specifically 
        for the 'IDT' method.
    """
    if method == 'IDT':
        method_kwargs = tm_kwargs.copy()
        if 'dNTPs' in method_kwargs:
            method_kwargs.pop('dNTPs')
        if 'DMSO' in method_kwargs:
            method_kwargs.pop('DMSO')
        t_anneal = (0.3 * min(mts) 
                    + 0.7 * mt.chem_correction(mt.Tm_GC(seq, 
                                                        valueset=7, 
                                                        **method_kwargs), 
                                                DMSO=tm_kwargs['DMSO']) - 14.9)
    else:  # Phusion method
        t_anneal = min(mts)
        if all(len(p) > 20 for p in seq):
            t_anneal += 3
    return round(t_anneal, 2)


def check_dimer(seq1: str, 
                seq2: str, 
                dict_format: bool = False,
                basepair: dict = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'}
                ) -> Union[str, dict]:
    """ 
    Check the dimerization of two sequences and return the best dimer found. 
    The dimerization is checked in an extremely simple way, by aligning the 
    two sequences and checking for WC basepairing at the same index.
    If dict_format is True, return a dictionary with all the dimers found. 

    Parameters:
    -----------
    seq1 (str): 
        The first sequence to check.
    seq2 (str): 
        The second sequence to check.
    dict_format (bool): 
        If True, return a dictionary with all the dimers found.
    basepair (dict):
        A dictionary with the basepairing rules. 
        Default is {'A': 'U', 'U': 'A', 'C': 'G', 'G': 'C'} for RNA.
        For DNA, use {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'}.

    Returns:
    --------
    str or dict:
        If dict_format is True, return a dictionary with the dimers found.
        If dict_format is False, return the best dimer found as a string.
    """
    dimers_dict = {}
    basepair = str.maketrans(basepair)
    top_seq = ' ' * (len(seq2) - 1) + str(seq1) + ' ' * len(seq2)
    bottom_seq = str(seq2[::-1])
    max_match = [top_seq, ' ' * len(top_seq), bottom_seq]

    for start_ind in range(len(seq1) + len(seq2) - 2):
        bp = ''
        for ind, b in enumerate(bottom_seq):
            if (b.translate(basepair) == top_seq[ind] 
                    and b != ' ' and top_seq[ind] != ' '):
                bp += '┊'
            else:
                bp += ' '
        num_bp = bp.count('┊')
        index_num = f"[{max(0, start_ind - len(seq2) + 1)}]"
        if num_bp > max_match[1].count('┊'):
            max_match = [index_num 
                         + top_seq, ' ' * len(index_num) 
                         + bp, ' ' * len(index_num) + bottom_seq]
        if dict_format:
            if num_bp not in dimers_dict:
                dimers_dict[num_bp] = []
            dimers_dict[num_bp].append('\n'.join(max_match))
        top_seq = top_seq[1:]

    return dimers_dict if dict_format else '\n'.join(max_match)


def auto_design_primers(seq: Seq, 
                        target_temp: float = 65.0, 
                        tm_func: Callable = make_tm_calculator()
                        ) -> Tuple[List[str], List[float]]:
    """
    Automatically designs forward and reverse primers for a given DNA sequence,
      optimizing for melting temperature (Tm) and primer quality.

    Parameters
    ----------
    seq : Seq
        The DNA sequence for which primers are to be designed.
    target_temp : float, optional
        The target melting temperature (Tm) for the primers, by default 65.0°C.
    tm_func : Callable, optional
        A function to calculate the melting temperature of a primer sequence, 
        by default uses `make_tm_calculator()`.

    Returns
    -------
    final_primers : list of str
        A list containing the best forward and reverse primer sequences.
    final_mts : list of float
        A list containing the melting temperatures (Tm) of the selected primers.

    Notes
    -----
    - The function evaluates primer candidates based on Tm, GC content, length, 
        and dimerization potential.
    - Returns empty string and 0 Tm if no suitable primer is found for a direction.
    """
    final_primers = []
    final_mts = []

    for direction in (1, -1):
        prim_length = 10
        primers_info = []

        to_prime = seq if direction == 1 else str(Seq(seq).reverse_complement())

        while prim_length < len(seq):
            primer = to_prime[:prim_length]
            tm = round(tm_func(primer), 2)

            if tm > (target_temp + 2.5):
                break
            if tm < (target_temp - 2.5):
                prim_length += 1
                continue

            score = 0
            if primer[-1] in 'GC':
                score += 1
            if primer[-2] in 'GC':
                score += 1
            if 18 < prim_length < 30:
                score += 1
            elif prim_length < 18:
                score -= 17 - prim_length
            if 40 < gc_fraction(primer, ambiguous='ignore') * 100 < 60:
                score += 1
            score -= abs(tm - target_temp) / 2

            dimer_score = max(check_dimer(primer, primer, dict_format=True).keys())
            score -= dimer_score / (len(primer) // 2)

            primers_info.append((score, tm, str(primer)))
            prim_length += 1

        primers_info.sort(reverse=True)
        if primers_info:
            final_mts.append(primers_info[0][1])
            final_primers.append(primers_info[0][2])
        else:
            final_primers.append("")
            final_mts.append(0)

    return final_primers, final_mts



"""
This module contains general utility functions and constants for the pyfurnace package.

It includes: 
functions for manipulating RNA strands and sequences, 
utilities for RNA structures (dot-bracket notation, pair maps, trees)
folding barriers calculation for canonical co-transcriptional RNA Origami
"""

import random
import warnings
from typing import Optional, Tuple, Dict, List, Union, Iterable
from .basepair import BasePair

###
### USEFUL SETS AND DICTIONARIES
###

#: Set of all accepted nucleotides, including standard bases and extended IUPAC codes.
nucleotides = {"A", 
               "U", 
               "C", 
               "G", # standard nucleotides
               "W", # A or U
               "M", # A or C
               "R", # A or G
               "Y", # C or U
               "K", # G or U
               "S", # G or C  
               "D", # A or G or U
               "H", # A or C or U
               "V", # A or C or G
               "B", # C or G or U
               "N", # any nucleotide
               "X", # any nucleotide for external Kissing Loops in ROAD
               '&', # separator
               }

#: Dictionary mapping IUPAC codes to corresponding sets of nucleotides.
#: Source: https://www.bioinformatics.org/sms/iupac.html
iupac_code = {"A": {"A"}, 
              "U": {"U"}, 
              "C": {"C"}, 
              "G": {"G"}, # standard nucleotides
              "W": {"A", "U"}, # A or U
              "M": {"A", "C"}, # A or C
              "R": {"A", "G"}, # A or G
              "Y": {"C", "U"}, # C or U
              "K": {"G", "U"}, # G or U
              "S": {"G", "C"}, # G or C
              "D": {"A", "G", "U"}, # A or G or U
              "H": {"A", "C", "U"}, # A or C or U
              "V": {"A", "C", "G"}, # A or C or G
              "B": {"C", "G", "U"}, # C or G or U
              "N": {"A", "U", "C", "G"}, # any nucleotide
              "X": {"A", "U", "C", "G"}, # any nucleotide for external Kissing Loops in ROAD
              "&": {"&"}, # separator
              }

#: Dictionary mapping IUPAC codes to the set of codes they can base pair with.
base_pairing = {"A": {"B", "D", "H", "K", "N", "U", "W", "Y"},
                "U": {"A", "B", "D", "G", "H", "K", "M", "N", "R", "S", "V", "W"},
                "C": {"B", "D", "G", "K", "N", "R", "S", "V"},
                "G": {"B", "C", "D", "H", "K", "M", "N", "S", "U", "V", "W", "Y"},
                "W": {"A", "B", "D", "G", "H", "K", "M", "N", "R", "S", "U", "V", "W",
                      "Y"},
                "M": {"B", "D", "G", "H", "K", "N", "R", "S", "U", "V", "W", "Y"},
                "R": {"B", "C", "D", "H", "K", "M", "N", "S", "U", "V", "W", "Y"},
                "Y": {"A", "B", "D", "G", "H", "K", "M", "N", "R", "S", "V", "W"},
                "K": {"A", "B", "C", "D", "G", "H", "K", "M", "N", "R", "S", "U", "V",
                      "W", "Y"},
                "S": {"B", "C", "D", "G", "H", "K", "M", "N", "R", "S", "U", "V", "W",
                      "Y"},
                "D": {"A", "B", "C", "D", "G", "H", "K", "M", "N", "R", "S", "U", "V",
                      "W", "Y"},
                "H": {"A", "B", "D", "G", "H", "K", "M", "N", "R", "S", "U", "V", "W",
                      "Y"},
                "V": {"B", "C", "D", "G", "H", "K", "M", "N", "R", "S", "U", "V", "W",
                      "Y"},
                "B": {"A", "B", "C", "D", "G", "H", "K", "M", "N", "R", "S", "U", "V",
                      "W", "Y"},
                "N": {"A", "B", "C", "D", "G", "H", "K", "M", "N", "R", "S", "U", "V",
                      "W", "Y"},
                # this symbol is used for external Kissing Loops in ROAD
                "X": {"A", "B", "C", "D", "G", "H", "K", "M", "N", "R", "S", "U", "V", 
                      "W", "Y", "X"},
                "&": {"&"} # separator
                }

#: Dictionary mapping dot-bracket symbols to their corresponding pairing symbols.
db_pairs = {"(": ")", "[": "]", "{": "}", "<": ">", "A": "a", "B": "b", "C": "c", 
            "D": "d", "E": "e", "F": "f", "G": "g", "H": "h", "I": "i", "J": "j",
            "K": "k", "L": "l", "M": "m", "N": "n", "O": "o", "P": "p", "Q": "q",
            "R": "r", "S": "s", "T": "t", "U": "u", "V": "v", "W": "w", "X": "x",
            "Y": "y", "Z": "z"}

#: Tuple containing all possible pseudoknot symbols used in dot-bracket notation.
all_pk_symbols = ('[', ']', '{', '}', '<', '>', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
                  'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V',
                  'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
                  'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x',
                 'y', 'z')

#: Set of all accepted symbols including nucleotides, dots, brackets, and special ROAD 
#: symbols.
accept_symbol = (
    nucleotides | {"." "(", ")"} | set(all_pk_symbols) | {
        "─", "│", "╭", "╮", "╰", "╯", "^", "*", "┼", "┊", '~', '◦',
        '↑', '↓', "⊗", "⊙", "▂", "▄", "█",
        "-", "|", "+", ":", "=", "!", " ", "/", "\\", "3", "5", "&"
    }
)

#: Set of symbols representing base pairs.
bp_symbols = {"┊", "=", ":", "!", "*"}

###
### STRING TRANSLATORS
###

#: String translator mapping all pseudoknot symbols to dots.
pseudo_to_dot = str.maketrans(''.join(all_pk_symbols), "." * len(all_pk_symbols))

#: String translator to map each dot-bracket opening symbol to its closing counterpart
#: and vice versa.
pair_db_sym = str.maketrans(
    ''.join(db_pairs.keys()) + ''.join(db_pairs.values()),
    ''.join(db_pairs.values()) + ''.join(db_pairs.keys())
)

#: String translator that removes all nucleotides from a string.
nucl_to_none = str.maketrans("", "", "".join(nucleotides))

#: String translator that removes all accepted symbols from a string.
symb_to_none = str.maketrans("", "", "".join(accept_symbol))

#: String translator that maps nucleotides (including IUPAC codes) to their  
#: complementary base.
nucl_to_pair = str.maketrans("AUCGWMRYKSDHVBNX", "UAGCWKYRKSHDBVNX")

#: String translator that removes all non-nucleotide symbols from a string.
only_nucl = str.maketrans("", "", "".join(accept_symbol - nucleotides))

#: String translator for horizontal flip of strand turning symbols.
horiz_flip = str.maketrans("╭╮╰╯/\\", "╮╭╯╰\\/")

#: String translator for vertical flip of strand turning symbols.
verti_flip = str.maketrans("╭╮╰╯/\\", "╰╯╭╮\\/")

#: String translator for 90 degree rotation of strand symbols.
rotate_90 = str.maketrans("╭╮╰╯│|─-", "╮╯╭╰──││")

#: String translator for converting normal symbols to ROAD strand symbols.
symb_to_road = str.maketrans("-|+=:*!", "─│┼┊┊┊┊")

###
### SEQUENCE FUNCTIONS
###

def complement(sequence: str) -> str:
    """
    Return the complement of an RNA sequence using IUPAC base-pairing rules.

    Parameters
    ----------
    sequence : str
        RNA sequence composed of standard and extended IUPAC nucleotide symbols.

    Returns
    -------
    str
        Complementary sequence with each base replaced by its pair.
    """
    return sequence.translate(nucl_to_pair)

def reverse_complement(sequence: str) -> str:
    """
    Return the reverse complement of an RNA sequence.

    Parameters
    ----------
    sequence : str
        RNA sequence to reverse and complement.

    Returns
    -------
    str
        The reverse-complemented sequence.
    """
    return sequence.translate(nucl_to_pair)[::-1]

def pair_nucleotide(nucleotide: str, iupac_symbol: str = 'N') -> str:
    """
    Return a nucleotide that can pair with the given nucleotide, optionally 
    constrained by an IUPAC symbol.

    Parameters
    ----------
    nucleotide : str
        The base to find a pair for.
    iupac_symbol : str, optional
        Constraint based on allowed base-pairing (IUPAC code).

    Returns
    -------
    str
        A valid pairing nucleotide, or the complement of the input base.
    """
    if iupac_symbol in 'AUCG':
        return iupac_symbol
    if iupac_symbol == 'K':
        if nucleotide == 'G':
            return 'U'
        elif nucleotide == 'U':
            return 'G'
    return nucleotide.translate(nucl_to_pair)

def mutate_nucleotide(sequence: str,
                      sequence_constraints: str,
                      nucl_ind: int,
                      pair_map: Dict[int, Optional[int]]
                      ) -> Tuple[Optional[str], Optional[str]]:
    """
    Mutate a nucleotide in the sequence and update its paired counterpart accordingly.

    Parameters
    ----------
    sequence : str
        Original RNA sequence.
    sequence_constraints : str
        A string of IUPAC codes representing base constraints for each position.
    nucl_ind : int
        Index of the nucleotide to mutate.
    pair_map : dict of int to int or None
        A mapping of indices representing base pairs.

    Returns
    -------
    tuple of (str or None, str or None)
        A tuple with the new nucleotide and its paired base if applicable.
    """
    new_set = iupac_code[sequence_constraints[nucl_ind]] - {sequence[nucl_ind]}
    if not new_set:
        return None, None
    new_nucl = random.choice(list(new_set))

    # get the index of the paired nucleotide
    paired_nucl = pair_map[nucl_ind] 

    # if the nucleotide is paired, get the paired nucleotide
    if paired_nucl is not None: 
        # get the possible nucleotides that can pair with the new nucleotide
        paired_nucl = pair_nucleotide(new_nucl, sequence_constraints[paired_nucl])
    return new_nucl, paired_nucl

def gc_content(seq: str, extended_alphabet: bool = True) -> float:
    """
    Calculate GC content (optionally including ambiguous nucleotides).

    Parameters
    ----------
    seq : str
        The RNA sequence.
    extended_alphabet : bool, optional
        Whether to include ambiguous IUPAC symbols (e.g., S, K) as partial GC.

    Returns
    -------
    float
        Proportion of GC content in the sequence.
    """
    total_count = len(seq)
    if not total_count:
        return 0
    gc_count = seq.count('G') + seq.count('C') 
    if extended_alphabet:
        gc_count += (seq.count('S') 
                     + sum(map(seq.count, ['M', 'R', 'Y', 'K', 'N', 'X'])) / 2 
                     + sum(map(seq.count, ['D', 'H'])) / 3 
                     + sum(map(seq.count, ['V', 'B'])) * 2 / 3 )
    percentage = (gc_count / total_count)
    return percentage

###
### STRUCTURE FUNCTIONS
###

def rotate_dot_bracket(dot_bracket: str, shift_left: int) -> str:
    """
    Rotate a dot-bracket structure leftward by a given number of positions.

    Parameters
    ----------
    dot_bracket : str
        RNA secondary structure in dot-bracket notation.
    shift_left : int
        Number of positions to rotate to the left.

    Returns
    -------
    str
        Rotated dot-bracket structure.
    """
    n = len(dot_bracket)
    
    # Step 1: Get the pair map
    pair_map = dot_bracket_to_pair_map(dot_bracket)
    
    # Step 3: Adjust the pair map
    new_pair_map = BasePair()
    for i, j in pair_map.items():
        new_i = (i - shift_left) % n
        new_j = (j - shift_left) % n if j is not None else None
            
        new_pair_map[new_i] = new_j
    
    return pair_map_to_dot_bracket(new_pair_map, n)
    
def dot_bracket_to_pair_map(dot_bracket: str) -> BasePair:
    """
    Convert dot-bracket notation into a pair map.

    Parameters
    ----------
    dot_bracket : str
        Secondary structure in dot-bracket notation.

    Returns
    -------
    BasePair
        A bidirectional dictionary mapping nucleotide indices to their paired 
        partner (or None).
    """
    pair_map = BasePair()
    stacks = [[] for _ in range(len(db_pairs))]
    for i, sym in enumerate(dot_bracket):
        if sym in db_pairs.keys():
            stack_num = list(db_pairs.keys()).index(sym)
            stacks[stack_num].append(i)
        elif sym in db_pairs.values():
            stack_num = list(db_pairs.values()).index(sym)
            if stacks[stack_num]: # pop the last element if the stack is not empty
                j = stacks[stack_num].pop()
                pair_map[j] = i
        else: # unpaired nucleotide: No pair
            pair_map[i] = None

    # If there are still elements that are not in the dictionary, add them as unpaired
    for i in range(len(dot_bracket)):
        if i not in pair_map:
            pair_map[i] = None

    return pair_map

def pair_map_to_dot_bracket(pair_map: Dict[int, Optional[int]],
                            structure_length: Optional[int] = None) -> str:
    """
    Convert a base pair map into dot-bracket notation.

    Parameters
    ----------
    pair_map : dict
        Mapping from nucleotide index to paired index or None.
    structure_length : int, optional
        Total length of the structure. If None, computed from the map.

    Returns
    -------
    str
        Dot-bracket notation representing the structure.
    """
    if structure_length is None:
        structure_length = max(pair_map.keys(), default=-1) + 1

    ### CREATE THE DOT BRACKET NOTATION ###
    done_pairs = set()

    # Prepare the variables
    dotbracket = ['.'] * structure_length 
    bracket_count = 0
    recheck = True
    stack = []

    while recheck: # recheck the structure, for pseudoknots
        recheck = False

        for i in range(structure_length):

            paired = pair_map[i]
            if paired is None or i in done_pairs or paired in done_pairs:
                continue # ignore unpaired positions or already paired positions

            if paired > i: # the current position will close the pair later
                
                # may be a pseudknot:  Check for clash
                if stack and paired > stack[-1]:
                    recheck = True # clash detected: recheck the structure later
                    continue
                else: # no clash: add the pair to the stack
                    stack.append(paired)
                    dotbracket[i] = list(db_pairs.keys())[bracket_count]
                    dotbracket[paired] = list(db_pairs.values())[bracket_count]
                    
            else: # paired already analyzed
                if not stack: # no pair in the stack
                    continue
                if i == stack[-1]: # the current position closes the last pair
                    # Remove the pair from the map
                    done_pairs.add(i)
                    done_pairs.add(paired)
                    stack.pop()

        stack = [] # reset the stack count
        bracket_count += 1

        if bracket_count >= 30:
            warnings.warn(f"Warning: Too many bracket types needed in to write the "
                          f"structure. Stopping at 30, 'Z' to 'z'.", stacklevel=3)
            break

    return ''.join(dotbracket)

def dot_bracket_to_stacks(dot_bracket: str,
                          only_opening: bool = False
                          ) -> Tuple[str, List[Tuple[int, int]]]:
    """
    Identify contiguous structural elements (stacks) in a dot-bracket string.

    Parameters
    ----------
    dot_bracket : str
        RNA structure in dot-bracket format.
    only_opening : bool, optional
        Whether to return only stacks that start with an opening symbol.

    Returns
    -------
    tuple of str and list of tuple
        - A reduced dot-bracket string marking distinct stacks.
        - A list of (start_index, end_index) tuples for each stack.
    """
    pair_map = dot_bracket_to_pair_map(dot_bracket)
    # a list of tuples containing starting index and the length of the stack
    stacks = [] 
    # a list of the reduced dot bracket symbols, that will be returned as a string
    reduced_dot_bracket = [] 

    stack_start = 0
    current_pair_sym = dot_bracket[0]
    last_stack_pair = pair_map[0] + 1 if pair_map[0] else None

    ### NOT NECESSARY 
    # open_pair_sym = '.' + ''.join(db_pairs.keys())

    for i, symbol in enumerate(dot_bracket + '_'):

        ### NOT NECESSARY 
        # # ignore the symbols that are not '.' or not opening pairs, 
        # # they should be already taken into accound
        # if symbol in open_pair_sym:
        #     continue

        # get the paired index of the current symbol
        stack_pair = pair_map.get(i) 

        # if symbol change, or the pairing is not consecutive, 
        # add the last stack to the list

        if (symbol != current_pair_sym 
                or stack_pair is not None 
                and stack_pair != last_stack_pair - 1):
            
            add_last_stack = True # add the last stack to the list by default
            # if only the opening stacks are needed, 
            # check if the last stack is an opening stack or unpaired

            if only_opening: 
                add_last_stack = (current_pair_sym in db_pairs.keys() 
                                  or current_pair_sym == '.')

            if add_last_stack:
                reduced_dot_bracket.append(current_pair_sym)
                stacks.append((stack_start, i - 1))
            stack_start = i

        last_stack_pair = stack_pair
        current_pair_sym = symbol

    return ''.join(reduced_dot_bracket), stacks


class Node():
    """
    A tree node representing a region of an RNA secondary structure.

    Attributes
    ----------
    index : int or None
        Position of the nucleotide in the sequence.
    label : str
        Structure symbol (e.g., '(', ')', '.', '&').
    paired_index : int or None
        Index of the nucleotide this node pairs with, if any.
    parent : Node or None
        Parent node in the tree.
    seq : str or None
        Sequence constraint at this node.
    children : list of Node
        Child nodes representing nested structures.
    """

    def __init__(self,
                 index: Optional[int] = None,
                 label: str = "5'",
                 parent: Optional['Node'] = None,
                 seq: Optional[str] = None,
                 **kwargs) -> None:
        """
        Initialize a Node object.
        Parameters
        ----------
        index : int or None
            Index of the nucleotide in the sequence.
        label : str
            Structure symbol (e.g., '(', ')', '.', '&').
        parent : Node or None
            Parent node in the tree.
        seq : str or None
            Sequence constraint at this node.
        kwargs : dict
            Additional attributes to set on the node.
        """
        # Initialize the node with the given parameters
        self.index = index
        self.label = label
        self.paired_index = None
        self.parent = parent
        self.seq = seq
        self.children: List[Node] = []
        if parent is not None:
            parent.children.append(self)
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self, level=0):
        """ 
        Return a string representation of the node and its children.
        """
        paired_str = ''
        if self.paired_index is not None:
            paired_str = f' -> {self.paired_index}'

        if level == 0:
            ret = f"{self.label}\n"
        else:
            ret = (f"╰{'──' * level} {self.label} - {self.seq} "
                   f"- {self.index} {paired_str} \n" )
        for child in self.children:
            ret += child.__repr__(level + 1)
        return ret

    def __str__(self):
        """
        Return a string representation of the node.
        """
        return self.__repr__()

    def search(self,
               target_index: Optional[int],
               target_label: Optional[str] = None) -> Optional['Node']:
        """
        Recursively search for a node with a matching index and label.

        Parameters
        ----------
        target_index : int or None
            Index to match.
        target_label : str or None
            Optional label to also match.

        Returns
        -------
        Node or None
            The first matching node found, or None if not found.
        """
        # Check if the current node matches the target criteria
        if (target_index == self.index 
                and (target_label is None 
                     or target_label == self.label)):
            return self
        elif (self.index is None 
                and self.parent is None 
                and abs(target_index) == float('inf')):
            return self # special case for the root node

        # Recursively search in child nodes
        for child in self.children:
            result = child.search(target_index=target_index, 
                                  target_label=target_label)
            if result:
                return result  # Return the node if found
    
        return None  # Return None if not found

def dot_bracket_to_tree(dot_bracket: str,
                        sequence: Optional[str] = None) -> Node:
    """
    Convert a dot-bracket RNA structure into a hierarchical tree of nodes.

    Parameters
    ----------
    dot_bracket : str
        RNA structure in dot-bracket format. Pseudoknots are ignored.
    sequence : str, optional
        Sequence constraints associated with each index.

    Returns
    -------
    Node
        Root node of the tree representing the full structure.
    """
    # Remove pseudoknots from the dot bracket
    dot_bracket = dot_bracket.translate(pseudo_to_dot)
    root = Node()
    current_node = root
    for i, label in enumerate(dot_bracket):
        if sequence is not None:
            seq_const = sequence[i]
        else:
            seq_const = None
        if label == "(": # add a new node, move to the new node
            new_node = Node(i, label, current_node, seq = seq_const)
            current_node = new_node
        # add the paired index to the parent, move to the parent node
        elif label == ")": 
            current_node.paired_index = i
            current_node = current_node.parent
        elif label == ".": # add a children to the current node, do not move
            new_node = Node(i, label, current_node, seq = seq_const)
        elif label == '&':
             new_node = Node(None, label, current_node)
    return root

def tree_to_dot_bracket(node: Node,
                        dot_bracket: Optional[List[str]] = None,
                        seq_constraints: Union[bool, List[str]] = False
                        ) -> Union[str, Tuple[str, str]]:
    """
    Convert a tree of Nodes into a dot-bracket structure and optional sequence 
    constraints.

    Parameters
    ----------
    node : Node
        Root node of the RNA structure tree.
    dot_bracket : list of str, optional
        Accumulator for the dot-bracket string.
    seq_constraints : bool or list of str
        If True, collect sequence constraints using the `seq` attribute.

    Returns
    -------
    str or tuple of (str, str)
        Dot-bracket string, and sequence constraints if requested.
    """

    if isinstance(seq_constraints, bool) and seq_constraints:
        # initialize the sequence constraints only if it's a bool
        seq_constraints = ["N"] # assume the sequence is at least one nucleotide long

    if dot_bracket is None:
        dot_bracket = ["."] # assume the structure is at least one nucleotide long

    if node.parent is not None: # don't add the root node
        if node.index >= len(dot_bracket):
            add_length = node.index - len(dot_bracket) + 1
            dot_bracket += ["."] * add_length
            if seq_constraints:
                seq_constraints += ["N"] * add_length

        dot_bracket[node.index] = node.label
        if seq_constraints and node.seq:
            seq_constraints[node.index] = node.seq

        if node.paired_index is not None:

            if node.paired_index >= len(dot_bracket):
                add_length = node.paired_index - len(dot_bracket) + 1
                dot_bracket += "." * add_length
                if seq_constraints:
                    seq_constraints += ["N"] * add_length

            dot_bracket[node.paired_index] = ")"
            if seq_constraints and node.seq:
                seq_constraints[node.paired_index] = node.seq.translate(nucl_to_pair)

    for child in node.children: # recursively add the children
        tree_to_dot_bracket(child, dot_bracket, seq_constraints)
    
    if node.parent is None: # we reached back the root node
        if seq_constraints:
            return ''.join(dot_bracket), ''.join(seq_constraints)
        
        return ''.join(dot_bracket)

###
### DISTANCE FUNCTIONS
###

def hamming_distance(s1: str, 
                     s2: str, 
                     ignore_ind: Iterable[int] = (),
                     **kwargs) -> int:
    """
    Compute the Hamming distance between two strings, ignoring specified indices.

    Parameters
    ----------
    s1 : str
        First sequence or structure.
    s2 : str
        Second sequence or structure.
    ignore_ind : iterable of int, optional
        Indices to exclude from comparison.

    Returns
    -------
    int
        Number of differing characters at unignored positions.
    """
    return sum((1 for i, (x, y) in enumerate(zip(s1, s2)) 
                    if x != y and i not in ignore_ind))

def base_pair_difference(s1: str,
                         s2: str,
                         pair_map1: Optional[Dict[int, Optional[int]]] = None,
                         pair_map2: Optional[Dict[int, Optional[int]]] = None,
                         ignore_ind: Iterable[int] = (),
                         accept_unpaired_ind: Iterable[int] = (),
                         **kwargs) -> List[int]:
    """
    Find indices where base pairing differs between two structures.

    Parameters
    ----------
    s1 : str
        First structure (dot-bracket notation).
    s2 : str
        Second structure.
    pair_map1 : dict, optional
        Precomputed pair map for s1.
    pair_map2 : dict, optional
        Precomputed pair map for s2.
    ignore_ind : iterable of int, optional
        Indices to ignore in the comparison.
    accept_unpaired_ind : iterable of int, optional
        Indices allowed to be unpaired in s2 without penalty.

    Returns
    -------
    list of int
        Indices of base pairs that differ.
    """
    if pair_map1 is None:
        pair_map1 = dot_bracket_to_pair_map(s1)
    if pair_map2 is None:
        pair_map2 = dot_bracket_to_pair_map(s2) 

    # Determine if an index should be considered based on the unpaired indices
    def check_ind(main_pair_map, i, accept_unpaired=False):
        if i in ignore_ind or main_pair_map[i] is None:
            return False
        elif accept_unpaired and i in accept_unpaired_ind:
            # Only count if paired in both but different
            return pair_map2[i] is not None and pair_map1[i] != pair_map2[i]  
        return pair_map1[i] != pair_map2[i]

    # pairs in s1 that are not in s2
    diff_set = [i for i in pair_map1 if check_ind(pair_map1, i, True)]  
    # pairs in s2 that are not in s1
    diff_set.extend(i for i in pair_map2 if check_ind(pair_map2, i))  
    
    return diff_set

def base_pair_distance(s1: str,
                       s2: str,
                       pair_map1: Optional[Dict[int, Optional[int]]] = None,
                       pair_map2: Optional[Dict[int, Optional[int]]] = None,
                       ignore_ind: Iterable[int] = (),
                       accept_unpaired_ind: Iterable[int] = (),
                       **kwargs) -> int:
    """
    Compute the number of base pair differences between two structures.

    Parameters
    ----------
    s1 : str
        First dot-bracket structure.
    s2 : str
        Second dot-bracket structure.
    pair_map1 : dict, optional
        Pair map for s1.
    pair_map2 : dict, optional
        Pair map for s2.
    ignore_ind : iterable of int, optional
        Indices to ignore.
    accept_unpaired_ind : iterable of int, optional
        Positions where being unpaired in s2 is allowed.

    Returns
    -------
    int
        Total number of mismatched base pairs.
    """
    return len(base_pair_difference(s1, 
                                    s2, 
                                    pair_map1, 
                                    pair_map2, 
                                    ignore_ind, 
                                    accept_unpaired_ind, 
                                    **kwargs))

###
### FOLDING BARRIERS
###

def folding_barriers(structure: str,
                     kl_delay: int = 150) -> Tuple[str, int]:
    """
    Compute the folding barriers for a given RNA secondary structure.
    This function is based on 
    ROAD: https://www.nature.com/articles/s41557-021-00679-1
    This function analyzes the dot-bracket representation of the RNA secondary 
    structure to determine folding barriers based on base pair topology and 
    kinetic delay.

    Parameters
    ----------
    structure : str
        The dot-bracket notation representing the secondary structure of the RNA. 
        If folding barriers is called from a Motif or Origami, the structure is 
        already provided by the object in the dot-bracket format.
    kl_delay : int, optional, default=150
        The number of nucleotides (nts) of delay before kissing loops (KLs) snap 
        closed. A typical realistic value is around 350 (~1 second), while a 
        more conservative  setting is 150 by default.

    Returns
    -------
    Tuple[str, int]
        A tuple containing:
        - A string representing the barrier map where:
        '─' = no barrier (0 penalty),
        '▂' = opening pair weak barrier (1 penalty),
        '▄' = cloding pair weak barrier (1 penalty),
        '█' = strong barrier (2 penalty).
        - An integer score indicating the total penalty based on barrier strengths.

    Notes
    -----
    The function assigns barrier strengths based on the topology of base pairs 
    and kinetic constraints, ensuring proper folding predictions.
    """
    structure = structure.replace('&', '')
    ss_len = len(structure)
    s_map = dot_bracket_to_pair_map(structure)
    barriers = [''] * len(structure)

    # don't count the nucleotides in the 3' terminal position
    terminal_3_bonus = 16

    current_barr_count = 0
    for ind in range(ss_len):
        # get the bracket at the position
        bra = structure[ind]
        # if bra == '&':
        #     barriers[ind] = '&'
        #     continue

        # set the default barrier to '─', no barrier
        barriers[ind] = '─'

        if ind > ss_len - terminal_3_bonus:
            if s_map[ind] is not None:
                barriers[s_map[ind]] = '─'
            continue

        ### unpaired nts are not barriers and reset the current barrier count
        if bra == '.':
            barriers[ind] = '─'
            current_barr_count = 0

        ### opening pairs may be barriers or not, 
        # indicate them with '▂', reset the barrier count
        elif bra == '(':
            barriers[ind] = '▂'
            current_barr_count = 0

        ### closing pairs may be barriers or not, 
        # depending on the topology count
        elif bra == ')':
            # if the opening pair is not blocked, 
            # we close and reset the current barrier count
            if barriers[s_map[ind]] == '▂':
                barriers[ind] = '─'
                barriers[s_map[ind]] = '─'
                current_barr_count = 0

            # if the closing pair is already blocked, then 
            # we mark the barrier reached and start counting
            elif barriers[s_map[ind]] == '█':
                # if the current barrier count is greater than 5, wa mark it with '█'
                if current_barr_count > 5:
                    barriers[ind] = '█'
                    barriers[s_map[ind]] = '▄'
                # if the current barrier count is less than 5, we mark it with '▄'
                else:
                    barriers[ind] = '▄'
                    barriers[s_map[ind]] = '▄'
                current_barr_count += 1

        # if the index is bigger than the kl_delay, 
        # we check if the internal KLs are closed
        if ind > kl_delay:
            
            # we check if the KLs are closed, 
            # if yes, we mark them with 'N'
            close_sym = structure[ind - kl_delay]
            if close_sym not in '(.)' and close_sym in db_pairs.values():
                barriers[ind - kl_delay] == '─'
                barriers[s_map[ind - kl_delay]] == '─'

                # for each nt in the delay, we check if there are any barriers, 
                # if yes, we mark them with '█'
                for k in range(s_map[ind - kl_delay], ind-kl_delay):
                    if barriers[k] == '▂':
                        barriers[k] = '█'

    penalty = {'█': 2, '▄': 1, '▂': 1, '─': 0}#, '&': 0}
    repl = [penalty[i] for i in barriers]
    score = sum(repl)
    return ''.join(barriers), score

###
### CUSTOM EXCEPTIONS
###


class MotifStructureError(Exception):
    """
    Exception raised when a motif's secondary structure is invalid or inconsistent.

    This error typically indicates issues such as:
    - Overlapping or conflicting strands
    - Joining strands with different directionalities
    - Logical inconsistencies in structural motifs
    """
    pass


class AmbiguosStructure(Warning):
    """
    Warning raised when a structure is ambiguous or potentially problematic.

    This can occur in cases where:
    - The strand routing is not clear

    Parameters
    ----------
    message : str
        Description of the ambiguity or issue.
    """
    def __init__(self, message: str) -> None:
        self.message: str = message

    def __str__(self) -> str:
        return repr(self.message)



import streamlit as st
from streamlit_option_menu import option_menu
import json
import os
import warnings

# App utilities
from utils import (load_logo, 
                   main_menu_style, 
                   copy_to_clipboard, 
                   write_format_text,
                   check_import_pyfurnace)
from utils.template_functions import symbols, reference, sanitize_input
from utils.design_functions import origami_build_view
check_import_pyfurnace()
from pyfurnace import prepare as prep

# Melting and primer logic
from pyfurnace.prepare.utils import (
    tm_methods, tm_models, default_values,
    calculate_gc, make_tm_calculator, check_dimer,
    auto_design_primers, annealing_temp
)


def upload_setting_button():
    """Allow to upload setting"""
    st.session_state['upload_setting'] = True
    return

def calculate_annealing(seq, mts, c_primer, nc_primer, tm_kwargs):

    col1, col2 = st.columns(2, gap='large')
    with col1:
        subcol1, subcol2 = st.columns([11,1], vertical_alignment='center')
        with subcol1:
            write_format_text(c_primer)
        with subcol2:
            copy_to_clipboard(c_primer, "")

        subcol1, subcol2, subcol3 = st.columns(3)
        with subcol1:
            st.markdown(f"GC content: {calculate_gc(c_primer)}%")
        with subcol2:
            st.markdown(f"**:green[Tm: {round(mts[0], 1)}°C]**")
        with subcol3:
            st.markdown(f"Length: {len(c_primer)}")

    with col2:
        
        subcol1, subcol2 = st.columns([11,1], vertical_alignment='center')
        with subcol1:
            write_format_text(nc_primer)
        with subcol2:
            copy_to_clipboard(nc_primer, "")

        subcol1, subcol2, subcol3 = st.columns(3)
        with subcol1:
            st.markdown(f"GC content: {calculate_gc(nc_primer)}%")
        with subcol2:
            st.markdown(f"**:green[Tm: {round(mts[1], 1)}°C]**")
        with subcol3:
            st.markdown(f"Length: {len(nc_primer)}")

    # show the primers preview, check the self-dimerization of the primer and 
    # check the dimer between the primers and the sequence
    c_bases = len(c_primer)
    nc_bases = len(nc_primer)
    with st.expander("Dimerization preview"):
        st.markdown(f'<div style="text-align: center;"><span style="color: #D00000">'
                    f'{c_primer}</span>{seq[c_bases]}[...]{seq[-nc_bases-1:]}</div>', 
                    unsafe_allow_html=True)
        st.markdown(f'<div style="text-align: center;">'
                    f'{prep.Seq(seq[:c_bases+1]).complement()}'
                    f'[...]'f'{prep.Seq(seq[-nc_bases - 1]).complement()}'
                    f'<span style="color: #D00000">'
                    f'{nc_primer[::-1]}</span></div>', 
                    unsafe_allow_html=True)
        st.divider()
        col1, col2 = st.columns(2, gap='large')
        with col1:
            self_dimer1 = check_dimer(c_primer, 
                                      c_primer, 
                                      basepair={'A': 'T', 
                                                'T': 'A', 
                                                'C': 'G', 
                                                'G': 'C'})
            write_format_text('Self-Dimer forward primer\n' + self_dimer1)

        with col2:
            self_dimer2 = check_dimer(nc_primer, 
                                      nc_primer, 
                                      basepair={'A': 'T', 
                                                'T': 'A', 
                                                'C': 'G', 
                                                'G': 'C'})
            write_format_text('Self-Dimer reverse primer\n' + self_dimer2)

        self_dimer12 = check_dimer(c_primer, 
                                   nc_primer, 
                                   basepair={'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'})
        write_format_text('Dimer between primers\n' + self_dimer12)

    # add a warning if the melting temperature of the primers is too different
    anneal_color = 'green'
    if abs(mts[0] - mts[1]) >= 5:
        st.warning('The difference of Tm should be below 5°C', icon="⚠️")
        anneal_color = 'red'
                    
    # take into account the two main method to calculate the PCR annealing 
    # temperature: IDT method and Phusion Buffer method
    t_anneal = annealing_temp(mts, 
                              seq, 
                              tm_kwargs, 
                              method=st.session_state.annealing_model.split()[0])
    if t_anneal < 50:
        st.warning('Is suggested a temperature of annealing higher than 50°C.', 
                   icon="⚠️")
        anneal_color = 'red'

    # write the annealing temperature in big
    st.markdown(f"### Anneal at: :{anneal_color}[{t_anneal}°C]")

def primers_tab(seq, mt_correct):
    """ Calculate the melting temperature of the primers and check 
    the dimer between the primers and the sequence"""

    # show the settings for the two primers: 
    # choose the number of bases and show the gc content and melting temperature
    st.write('\n')
    col1, col2 = st.columns(2, gap='large')
    mts = [0, 0]
    # settings for the coding primer
    with col1:
        with st.columns(5)[2]:
            st.markdown('###### Forward')

        c_bases = st.slider('Primer length:', 
                            min_value=5, max_value=50, 
                            value=st.session_state.prim_fwd_len,
                            key=f"coding_primer_{st.session_state.auto_count}")
        c_primer = seq[:c_bases]

        mts[0] = round(mt_correct(c_primer),2)

    # settings for the non-coding primer
    with col2:
        with st.columns(5)[2]:
            st.markdown('###### Reverse')

        nc_bases = st.slider('Primer length:', 
                             min_value=5, max_value=50, 
                             value=st.session_state.prim_rev_len, 
                             key=f"non_coding_primer_{st.session_state.auto_count}")
        
        nc_primer = str(prep.Seq(seq[-nc_bases:]).reverse_complement())

        mts[1] = round(mt_correct(nc_primer),2)

    return mts, c_primer, nc_primer

def tm_parameters():
    # load settings if you want to upload custom settings
    load_settings = st.checkbox("Upload previous energy parameter:",
                                help='If you have saved a json file with the '
                                    'energy parameters, you can upload it and'
                                    ' use the same settings.')
    if load_settings:
        saved_setting = st.file_uploader("Upload previous energy parameter "
                                            "(optional):",
                                            on_change=upload_setting_button,
                                        type='json',
                                        help='If you have saved a json file '
                                            'with the energy parameters, you '
                                            'can upload it and use the same '
                                            'settings.')
        
        # if load the settings, upload the valuse from the json file
        if saved_setting and st.session_state['upload_setting']:
            # To read file as bytes:
            session_old = json.load(saved_setting)
            for key, value in session_old.items():
                st.session_state[key] = value
            st.session_state['upload_setting'] = False
            st.rerun()


    # initialize session state values for the energy model
    for key, val in default_values.items():
        if key not in st.session_state:
            st.session_state[key] = val

    # select the energy model
    st.write('Energy model:')
    col1, col2, col3 = st.columns([2, 2, 1])

    # select the melting temperature method
    with col1: 
        mt_type = st.selectbox("Melting Temperature Method", 
                                list(tm_methods), 
                                help=(prep.mt.__doc__), 
                                key='mt_method', 
                                label_visibility='collapsed')
        method = tm_methods[mt_type]

    # select the melting temperature model
    with col2:
        mt_model = st.selectbox("Energy Correction", 
                                list(tm_models[mt_type]), 
                                key='mt_model', 
                                label_visibility='collapsed')
    
    # add a button to show the help of the selected method
    with col3:
        help_model = st.button("Show info", key='energy_model_info')
    if help_model:
        st.help(method)
        
    # buffer correction parameters
    st.divider()
    if method != list(tm_methods)[2]:

        st.write('Buffer corrections (mM):')

        cols = st.columns(8)
        with cols[0]: 
            na = st.number_input('Na', key="Na")
        with cols[1]: 
            k = st.number_input('K', key='K')
        with cols[2]: 
            tris = st.number_input('Tris', key='Tris')
        with cols[3]: 
            mg = st.number_input('Mg', 
                                value=st.session_state["Mg"], 
                                key='Mg')
        with cols[4]: 
            dntps = st.number_input('dNTPs', 
                                        value=st.session_state["dNTPs"],
                                        key='dNTPs')
        with cols[5]: 
            dmso = st.number_input('DMSO (%)', 
                                    value=st.session_state["DMSO (%)"],
                                    key='DMSO (%)')
        with cols[6]: 
            correction_met = st.selectbox('Method', 
                                            [0, 1, 2, 3, 4, 6, 7],
                                            key='Method')
        with cols[7]: 
            help_correction = st.button("Show info", 
                                        key='energy_correction_info')

        tm_kwargs = {'Na': na, 
                        'K': k, 
                        'Tris': tris, 
                        'Mg': mg, 
                        'saltcorr': correction_met,
                        'dNTPs': dntps, 
                        'DMSO': dmso}

    if help_correction:
        # add a button to show the help of the selected method
        st.help(prep.mt.salt_correction)
        st.help(prep.mt.chem_correction)

    primer_conc = st.number_input('Primer conc. (nM)', key="Primer")

    # create the function to calculate the TM
    calculate_mt = make_tm_calculator(method_name=mt_type,
                                        model_name=mt_model,
                                        primer_conc=primer_conc,
                                        tm_kwargs=tm_kwargs)

    # save settings and allow download
    with open('energy_parameters.json', 'w') as settings:
        # cannot save session state as it is, 
        # I have to convert it to a dictionary
        session_dict = {key: st.session_state[key] for key in default_values}
        json.dump(session_dict, settings)
    with open("energy_parameters.json", "rb") as file:
        btn = st.download_button(
            label="Download energy parameters",
            data=file,
            file_name="energy_parameters.json",
            mime="application/json",
            help='Save the current settings (e.g. ions concentration, '
                    'Tm model), so you can easily reload them if you refresh '
                    'the page!'
            )
        
    return calculate_mt, tm_kwargs

def primers_setup():
    if "dna_template" not in st.session_state:
        st.session_state["dna_template"] = ''

    # take the input sequence and sanitize it
    seq = sanitize_input(st.text_input("Input sequence:", 
                                       value = st.session_state["dna_template"]))

    # check the symbols in the sequence
    if set(seq) - symbols:
        st.warning('The sequence contains symbols not included in the '
                    '[IUPAC alphabet](https://www.bioinformatics.org/sms/iupac.html).',
                    icon="⚠️")
    if 'U' in seq:
        st.error('The DNA template contains U', icon=":material/personal_injury:")
    
    if not seq:
        st.stop()
    elif len(seq) < 40:
        st.error('The sequence is too short', icon=":material/personal_injury:")
        st.stop()
    
    ###
    # Various for PCR settings
    ###

    mcol1, mcol2, mcol3, mcol4 = st.columns(4, 
                                            vertical_alignment='center',
                                            gap='large') 
    with mcol1:
        with st.popover("Melting temperature parameters",
                        use_container_width=True):
            
            calculate_mt, tm_kwargs = tm_parameters()
    
    with mcol2:
        with st.popover('Add restriction sites',
                        use_container_width=True):
            restric_site_1 = st.text_input("Restriction site sequence (5'->3')"
                                           " before the promoter:")
            restric_site_2 = st.text_input("Restriction site sequence (5'->3')"
                                           " after the fragment:")
            
    with mcol3:
        with st.popover('Annealing calculation',
                        use_container_width=True):
            st.selectbox("Calculate annealing:", 
                        ['IDT method [2]', 'Phusion method [3]'],
                        key='annealing_model',
                        label_visibility='collapsed',
                        help='Select the method to calculate the annealing '
                             'temperature. Check the references for more information.'
                             )
    with mcol4:
        # default primers length
        with st.popover('**Auto design primers**', 
                        icon=":material/precision_manufacturing:",
                        use_container_width=True):
            
            target_temp = st.number_input('Target melting temperature (°C)', 
                                value=65, min_value=0, max_value=100, step=1)
            if st.button("Auto design primers", 
                      key='auto_design_primers',
                      help='Automatically design the primers for the sequence. '
                           'The primers are designed to have a melting temperature '
                           'and follow best practices (see documentation).'):

                status = st.empty()
                status.info("Designing primers...", 
                            icon=":material/precision_manufacturing:")

                primers, final_mts = auto_design_primers(seq, 
                                                         target_temp, 
                                                         calculate_mt)

                st.session_state.auto_count += 1

                # set the new primers length in the session state
                if (st.session_state.prim_fwd_len != len(primers[0]) or 
                        st.session_state.prim_rev_len != len(primers[1])):
                    st.session_state.prim_fwd_len = len(primers[0])
                    st.session_state.prim_rev_len = len(primers[1])

                status.success(f"Primer designed!", 
                                icon=":material/precision_manufacturing:")
                

    seq = restric_site_1 + seq + str(prep.Seq(restric_site_2).reverse_complement())
    
    mts, c_primer, nc_primer = primers_tab(seq, 
                                           calculate_mt)

    calculate_annealing(seq, mts, c_primer, nc_primer, tm_kwargs)

    # add bibliography
    reference(True)


def md_setup():

    # check if the origami blueprint is good
    if (not st.session_state.get('origami') or
        not all(n in 'AUCG&' for n in st.session_state.origami.sequence)):

        st.error("The origami blueprint is empty or it doesn't contain a sequence",
                 icon=":material/personal_injury:")
        
        col1, col2 = st.columns(2)
        with col1:
            st.page_link("pages/1_Design.py", 
                         label="**:orange[Have a look to the blueprint]**", 
                         icon=":material/draw:")
        
        with col2:
            st.page_link("pages/2_Generate.py", 
                         label="**:orange[Generate the sequence]**", 
                         icon=":material/network_node:")
        st.stop()
    
    # show the origami blueprint
    with st.expander('**Origami preview:**', expanded=True):
        origami_build_view('Origami split view')

    st.divider()
    col1, col2 = st.columns(2)
    with col1:
        oxdna_dir = st.text_input('OxDNA directory '
                                  '(e.g. "~/Documents/software/oxDNA")')
    with col2:
        temp = st.number_input('Temperature (°C)', 
                               value=37, min_value=0, max_value=100, step=1)

    help_mc_rel = ("Use Monte Carlo relaxation to relax the origami before the MD "
                  "simulation, using external forces to keep the basepairing.")
    help_md_rel = ("Use Molecular Dynamics to relax the structure, using forces to"
                   "maintain the basepairing")
    help_md_equ = ("An extra Molecular Dynamics simulation to make sure the Origami"
                   "is fully relaxed and not bias the starting configuration of the " 
                   "production. **External forces for pseudoknots only** are used "
                   "in this simulation, to make sure the pseudoknots are paired "
                   "at the start of the production")
    help_md_run = ("The final Molecular Dynamics simulation, to simulate the "
                   "structure without any external force.")
    
    num_input = st.toggle("Use numerical input", key='num_input')

    col1, col2, col3, col4 = st.columns(4)
    with col1:
        if num_input:
            step_mc_rel = st.number_input('MC relaxation steps', 
                                          min_value=0,
                                          value=int(5e3),
                                          help=help_mc_rel,
                                          )
        else:
            step_mc_rel = st.slider('MC relaxation steps', 
                                    min_value=0,
                                    value=int(5e3),
                                    max_value=int(10e3),
                                    step=int(1e3),
                                    help=help_mc_rel,
                                    format="%0.0e"
                                    )
    with col2:
        if num_input:
            step_md_rel = st.number_input('MD relaxation steps', 
                                          min_value=0,
                                          value=int(1e7),
                                          help=help_md_rel,
                                          )
        else:
            step_md_rel = st.slider('MD relaxation steps', 
                                    min_value=0,
                                    value=int(1e7),
                                    max_value=int(1e8),
                                    step=int(1e6),
                                    help=help_md_rel,
                                    format="%0.0e"
                                    )
    with col3:
        if num_input:
            step_md_equ = st.number_input('MD equilibration steps', 
                                          min_value=0,
                                          value=int(1e8),
                                          help=help_md_equ,
                                          )
        else:
            step_md_equ = st.slider('MD equilibration steps', 
                                    min_value=0,
                                    value=int(1e8),
                                    max_value=int(1e9),
                                    step=int(1e7),
                                    help=help_md_equ,
                                    format="%0.0e"
                                    )
    with col4:
        if num_input:
            step_md_run = st.number_input('MD production steps', 
                                          min_value=0,
                                          value=int(1e9),
                                          help=help_md_run,
                                          )
        else:
            step_md_run = st.slider('MD production steps', 
                                    min_value=0,
                                    value=int(1e9),
                                    max_value=int(1e10),
                                    step=int(1e8),
                                    help=help_md_run,
                                    format="%0.0e"
                                    )
            
    if not oxdna_dir or not temp:
        st.stop()

    zip_path = prep.oxdna_simulations(origami=st.session_state.origami,
                                      oxdna_directory=oxdna_dir,
                                      temperature=temp,
                                      mc_relax_steps=step_mc_rel,
                                      md_relax_steps=step_md_rel,
                                      md_equil_steps=step_md_equ,
                                      md_prod_steps=step_md_run
                                      )
    # add a button to download the zip file
    st.divider()

    col1, col2 = st.columns(2, vertical_alignment='bottom')
    with col1:
        name = st.text_input('Name of the simulation:', value='origami_simulation')
    with col2:
        with open(zip_path, "rb") as file:
            st.download_button(
                label="Download MD simulation files",
                data=file,
                file_name=f"{name}.zip",
                mime="application/zip",
                help='Download the MD simulation files. '
                     'The zip file contains the input files for the MD simulation.'
                )
        
    # delete the zip file
    if os.path.exists(zip_path):
        os.remove(zip_path)
    return

if __name__ == "__main__":
    ### set the logo of the app
    load_logo()
    warnings.filterwarnings("ignore") # ignore warnings

    if "prepare_ind" not in st.session_state:
        st.session_state.prepare_ind = 0
    if 'prim_fwd_len' not in st.session_state:
        st.session_state.prim_fwd_len = 21
    if 'prim_rev_len' not in st.session_state:
        st.session_state.prim_rev_len = 21
    if 'auto_count' not in st.session_state:
        st.session_state.auto_count = 0

    # create the tabs with the functions
    st.header('Prepare', help='Design primers for your DNA template '
                               'or prepare the Origami for OxDNA simulation.')
    option_data = {'Primers': "bi bi-arrow-left-right",
                   'MD simulations': "bi bi-cpu"}

    selected_operation = option_menu(None, 
                                    list(option_data.keys()),
                                    icons=list(option_data.values()),
                                    menu_icon="cast", 
                                    orientation="horizontal",
                                    default_index=st.session_state.prepare_ind,
                                    styles=main_menu_style)
    
    if selected_operation == 'MD simulations':
        if st.session_state.prepare_ind == 0:
            st.session_state.prepare_ind = 1
            st.rerun()
        md_setup()
    elif selected_operation == 'Primers':
        if st.session_state.prepare_ind == 1:
            st.session_state.prepare_ind = 0
            st.rerun()
        primers_setup()
    


MOTIF_ICONS = {
    # Connection motifs
    'Tetraloop': "bi bi-braces",
    'start_end_stem': "bi bi-skip-end-fill",
    'stem_cap_link': "bi bi-align-start",
    'vertical_double_link': "bi bi-arrow-down-up",
    'vertical_link': "bi bi-arrow-up",
    
    # Structural motifs
    'Stem': "bi bi-layers",
    'Dovetail': "bi bi-fullscreen-exit",
    'LambdaTurn': "bi bi-sign-turn-slight-right",
    'ThreeWayJunction': "bi bi-sign-intersection-side",
    "Bend90": "bi bi-arrow-90deg-right",

    # Kissing loop motifs
    'KissingDimer': "bi bi-yin-yang",
    'KissingLoop180': "bi bi-square-half",
    'KissingLoop120': "bi bi-triangle-half",
    'BranchedKissingLoop': 'bi bi-align-top',
    'BranchedDimer': 'bi bi-tree',
    'KissingDimer120': 'bi bi-send',


    # Common aptamers
    "Broccoli": "bi bi-lightbulb",
    "Pepper": "bi bi-stoplights",
    "Ispinach": "bi bi-lamp",
    "MS2": "bi bi-signpost",
    "Streptavidin": "bi bi-link-45deg",

    None: "bi bi-bezier2",

}



import inspect
import streamlit as st
from streamlit import session_state as st_state

### pyFuRNAce modules
from utils import load_logo, check_import_pyfurnace
from utils.template_functions import sanitize_input
check_import_pyfurnace()
import pyfurnace as pf
import pyfurnace.design.utils.origami_lib as origami_lib


def quickstart():
    with st.popover(":red[New to pyFuRNAce? Start here!]",
                    use_container_width=True,):
        st.write("Do you have a sequence/structure (dot-bracket)?")
        start = st.radio("Select an option:", 
                         index=None,
                         options=["Yes", 
                                  "No"],
                         horizontal=True,
                         label_visibility="collapsed",
                         key="design_choice")
        
        if not start:
            return
        
        elif start == "Yes":
            st.write("Great! You can start by adding either your sequence, "
                     "your dot-bracket structure or both.")
            col1, col2 = st.columns(2)
            with col1:
                sequence = st.text_input("Sequence:", 
                                        key="Sequence", 
                            help="To generate an RNA origami from a sequence," 
                                 " we fold it with viennaRNA and then convert " 
                                 "the dot bracket notation to an Origami, "
                                 "as an collection of stems and bulges.") 

            with col2:
                structure = st.text_input("Dot-bracket structure:", 
                                          key="Dot-bracket_structure", 
                            help="Add a dot-bracket structure to generate "
                                " a blueprint. If the secondary structure "
                                " of known aptamers is detected, the aptamer "
                                " motif will be used in the Origami. ")
                
            structure = sanitize_input(structure)
            sequence = sanitize_input(sequence) 

            if not (structure or sequence):
                return
            try:
                pf.Origami.from_structure(structure=structure, 
                                          sequence=sequence)
            except ValueError as e:
                st.error(f"Error: {e}", icon=":material/personal_injury:")
            else:
                st.success("Origami created successfully!", 
                           icon=":material/check_circle:")
                
                st_state.code = ['import pyfurnace as pf\n'
                                         f'origami = pf.Origami.from_structure('
                                         f'structure="{structure}", '
                                         f'sequence="{sequence}")']
                st_state.origami = pf.Origami.from_structure(structure=structure,
                                                             sequence=sequence)
                
                st.page_link("pages/1_Design.py", 
                            label=":orange[Switch to the Design page]", 
                            icon=":material/draw:")
                
        elif start == "No":
            st.write("No problem! You can design from scratch, "
                     "or loading a template.")

            templ_funcs = {name.replace('template_', ''): func 
                                for name, func in vars(origami_lib).items()
                                    if callable(func) and name.startswith("template_")}

            template_docs = 'Available templates:\n\n'
            for name, func in templ_funcs.items():
                template_docs += f'**--> {name}** \n\n ' 
                docs = inspect.getdoc(func)
                template_docs += f'{docs.split("Returns")[0].strip()}\n\n'

            template = st.selectbox("Select a template:",
                                    options=list(templ_funcs.keys()),
                                    index=None,
                                    help=template_docs)
            if template:
                func_code = inspect.getsource(templ_funcs[template])
                # remove one layer of indentation
                only_code = func_code.split('"""')[2].split('return')[0]
                func_code = '\n'.join([line[4:] for line in only_code.splitlines()])

                st_state.origami = templ_funcs[template]()
                st_state.code = [f'{func_code.strip()}\n']
            
                st.success(f"Template {template} loaded successfully!", 
                           icon=":material/check_circle:")
                
            st.page_link("pages/1_Design.py", 
                        label=":orange[Switch to the Design page]", 
                        icon=":material/draw:")



if __name__ == '__main__':
    load_logo() 
    check_import_pyfurnace()

    st.write("# Hello and Welcome to pyFuRNAce!")

    st.write('Design and generate RNA nanostructures in few simple steps.')

    st.page_link("pages/1_Design.py", 
                 label=":orange[Design:]", 
                 icon=":material/draw:")

    st.markdown("- Design your RNA nanostructure and download it as "
                "textfile/python script.")

    st.page_link("pages/2_Generate.py", 
                 label=":orange[Generate:]", 
                 icon=":material/network_node:")

    st.markdown("- Generate the RNA sequence that matches the desired dot-bracket"
                " notation for the nanostructure.")

    st.page_link("pages/3_Convert.py", 
                 label=":orange[Convert:]", 
                 icon=":material/genetics:")

    st.markdown("- Prepare the DNA template for you RNA Origami, search subsequences"
                " and search for dimers.")

    st.page_link("pages/4_Prepare.py", 
                 label=":orange[Prepare:]", 
                 icon=":material/sync_alt:")

    st.markdown("- Design primers for your DNA template or prepare the Origami for "
                "OxDNA simulation.")
    
    
    _, col, _ = st.columns(3)
    with col:
        quickstart()
    
    st.divider()

    st.write("### About pyFuRNAce")
    
    st.markdown("pyFuRNAce is an open-source Python package and web-based design "
                "engine for creating complex RNA nanostructures using the "
                " co-transcriptional RNA origami approach.")
    st.markdown(" - **GitHub**: [Biophysical-Engineering-Group/pyFuRNAce]"
                "(https://github.com/Biophysical-Engineering-Group/pyFuRNAce)")
    st.markdown(" - **PyPI**: [pyfurnace](https://pypi.org/project/pyfurnace/)")
    st.markdown(" - **Documentation**: [Read the Docs]"
                "(https://pyfurnace.readthedocs.io/en/latest/)")
    st.markdown(" - bug reports, feature requests or any other questions, "
                "please reach out to us via the " 
                "[GitHub Issues](https://github.com/Biophysical-Engineering"
                "-Group/pyFuRNAce/issues)"
                " or the "
                "[GitHub Discussions](https://github.com/Biophysical-Engineering"
                "-Group/pyFuRNAce/discussions)."
                )
    
    st.divider()
    
    st.write("#### Check out the 1-min demo video:")
    st.video("https://github.com/Biophysical-Engineering-Group/pyFuRNAce/blob"
             "/main/vid/demo_1min.mp4?raw=true", 
             format="video/mp4", 
             start_time=0, 
             subtitles=None, 
             loop=True, 
             autoplay=True, 
             muted=True)